import unittest

import vtrace.breakpoints

# TODO: move testmods.hookbptest tests in here as well.

class MockTrace(object):
    def getCurrentThread(self):
        return 0

class HookBpPostTests(unittest.TestCase):
    def test_PostHitNoCallinfo(self):
        '''
        tests condition where:
        1. thread 1 hits hookbp and sets up postbp
        2. thread 2 hits postbp (there's nothing in callinfo for thread 2)
        '''
        hbp = vtrace.breakpoints.HookBreakpoint('0xdeadbeef')
        phbp = vtrace.breakpoints.PostHookBreakpoint('0xcafebabe', hbp)

        phbp.notify(None, MockTrace())



import types
import unittest
import vtrace.tests as vt_tests
from vtrace.lockstep import LockStepper



breakpoints = {
    ('windows', 'i386'): ('ntdll.RtlAllocateHeap', 500),
    ('windows', 'amd64'): ('ntdll.RtlAllocateHeap', 500),
    # 'linux': 'libc.exit',
    # 'freebsd': 'libc.exit',
}

class LockStepTest(vt_tests.VtraceProcessTest):

    def test_emu_lockstep(self):
        plat = self.trace.getMeta('Platform')
        arch = self.trace.getMeta('Architecture')

        stepinfo = breakpoints.get((plat, arch))
        if stepinfo is None:
            raise unittest.SkipTest('No Lockstep Break: %s %s' % (plat, arch))

        symname, stepcount = stepinfo
        untilva = self.trace.parseExpression(symname)

        # Just like runUntilExit()
        self.proc.stdin.write('testmod\n')
        self.proc.stdin.flush()
        self.trace.run(until=untilva)

        lock = LockStepper(self.trace)
        lock.stepi(stepcount)



import unittest

import vivisect
import vivisect.symboliks.archs.amd64 as sym_amd64

from vstruct.primitives import *
from vivisect.symboliks.common import *


class CallingConvTest(unittest.TestCase):
    def test_getSymbolikArgs(self):
        '''
        tests generating the symbolic definitions from the envi calling
        conventions.
        '''
        cc = sym_amd64.MSx64CallSym()

        argc = 5
        sargs = cc.getSymbolikArgs(None, 'a'*argc, update=False)
        assert(len(sargs) == argc)

        argc = 15
        sargs = cc.getSymbolikArgs(None, 'b'*argc, update=False)
        assert(len(sargs) == argc)

    def NEWPtest_setSymbolikArgs(self):
        '''
        tests setting and then getting the args for a function.  we manually
        smash in the function args for now *and* hardcode the test to a func
        in 64-bit pre-run win32k.sys.viv workspace.
        '''
        vw = vivisect.VivWorkspace()
        vw.loadWorkspace('win32k.sys.viv')
        fva = 0xfffff97fff1706a0
        args = [(v_uint64, None), ] * 6
        vw.setFunctionArgs(fva, args)

        ctx = sym_amd64.Amd64SymbolikAnalysisContext(vw)
        emu = ctx.getFunctionEmulator(fva)
        cc = sym_amd64.MSx64CallSym()

        argc = len(args)
        argv = [0x01, 0x02, 0x03, 0x04, 0x05, 0x06]
        cc.setSymbolikArgs(emu, argv)
        sargs = cc.getSymbolikArgs(emu, argv, update=False)
        for idx, arg in enumerate(sargs):

            val = arg.update(emu)
            if isinstance(arg, Mem):
                val = emu.readSymMemory(arg, 8)
                arg = arg.update(emu)
                arg = arg.reduce(emu)

        assert(len(sargs) == argc)



import os
import sys
import zipfile

'''
A utility package (in a central location) for packaging up
zip files full of source or whatever...
'''

def callback(z, dname, files):
    if dname.find('.svn') != -1:
        return

    for fname in files:
        if fname.endswith('.py'):
            fpath = os.path.join(dname, fname)
            z.write(fpath)

def addSource(z, dname):
    os.path.walk(dname, callback, z)

def main():

    zipname = sys.argv[1]
    pz = zipfile.PyZipFile(zipname, 'w')

    dirnames = sys.argv[2:]
    if not len(dirnames):
        dirnames = [ dname for dname in os.listdir('.') if os.path.isdir(dname) and dname != '.svn' ]

    for dirname in dirnames:
        addSource(pz, dirname)
        pz.writepy(dirname)

    pz.close()

if __name__ == '__main__':
    sys.exit(main())




"""
The new top level home for all the envi architecture modules.
"""
import os
import sys
import binascii

import envi.common as e_common


def dismain(d):
    '''
    Easy utility for implementing stand-alone disassembler utils...
    '''

    # TODO: this belongs as a standalone utility, not in an __init__ file
    if os.path.isfile(sys.argv[1]):
        with open(sys.argv[1], 'rb') as f:
            b = f.read()
    else:
        b = binascii.unhexlify(sys.argv[1])

    offset = 0
    va = 0x41414141
    while offset < len(b):
        op = d.disasm(b, offset, va+offset)
        print('0x%.8x %s %s' % (va+offset, e_common.hexify(b[offset:offset+len(op)]).ljust(16), repr(op)))
        offset += len(op)



import envi
import envi.cli as e_cli
import envi.common as e_common
import envi.archs.arm.regs as e_arm_regs
import envi.archs.thumb16.disasm as e_thumb

def armdis(db, line):
    '''
    Disassemble arm instructions from the given address.

    Usage: armdis <addr_exp>
    '''
    disasmobj = e_arm.ArmDisasm()
    armthumdis(db, line, disasmobj)

def thumbdis(db, line):
    '''
    Disassemble thumb instructions from the given address.

    Usage: thumbdis <addr_exp>
    '''
    disasmobj = e_thumb.ThumbDisasm()
    armthumdis(db, line, disasmobj)


def armthumbdis(db, line, disasmobj):
    '''
    Core of disassmbly, for code-reuse.  Only difference is the object actually
    doing the disassembly.
    '''
    t = db.getTrace()

    argv = e_cli.splitargs(line)
    size = 20
    argc = len(argv)
    if argc == 0:
        addr = t.getProgramCounter()
    else:
        addr = t.parseExpression(argv[0])

    if argc > 1:
        size = t.parseExpression(argv[1])

    bytez = t.readMemory(addr, size)
    offset = 0

    db.vprint("Dissassembly:")
    while offset < size:
        va = addr + offset
        op = disasmobj.disasm(bytez, offset, va)
        obytez = bytez[offset:offset+len(op)]

        db.canvas.addVaText('0x%.8x' % va, va=va)
        db.canvas.addText(": %s " % e_common.hexify(obytez).ljust(17))
        op.render(db.canvas)
        db.canvas.addText("\n")

        offset += len(op)

def togglethumb(db, line):
    '''
    Toggle Thumb Mode
    '''
    t = db.getTrace()
    cur_t = t.getRegister(e_arm_regs.REG_T)
    new_t = not cur_t
    arch = (envi.ARCH_ARMV7, envi.ARCH_THUMB)[new_t]
    t.setRegister(e_arm_regs.REG_T, new_t)

    db.canvas.addText("Toggled Thumb Mode: %r\n" % new_t)

def vdbExtension(vdb, trace):
    vdb.addCmdAlias('db', 'mem -F bytes')
    vdb.addCmdAlias('dw', 'mem -F u_int_16')
    vdb.addCmdAlias('dd', 'mem -F u_int_32')
    vdb.addCmdAlias('dq', 'mem -F u_int_64')
    vdb.addCmdAlias('dr', 'mem -F "Deref View"')
    vdb.addCmdAlias('ds', 'mem -F "Symbols View"')
    vdb.registerCmdExtension(armdis)
    vdb.registerCmdExtension(thumbdis)
    vdb.registerCmdExtension(togglethumb)



import sys
import json
import argparse

import vivisect
import vivisect.const as v_const


def setup():
    ap = argparse.ArgumentParser('Analyze a binary and dump certain features to json')
    ap.add_argument('path', help='Path to binary file to dump features')
    ap.add_argument('--output', default='features.json')
    return ap


def dumpFunctions(vw):
    ret = {}
    for fva in vw.getFunctions():
        # fva: (size, instruction count, code blocks
        size = vw.getFunctionMeta(fva, 'Size')
        inst = vw.getFunctionMeta(fva, 'InstructionCount')
        blks = vw.getFunctionBlocks(fva)

        ret[fva] = (size, inst, blks)

    return ret


def main(argv):
    opts = setup().parse_args(argv)

    vw = vivisect.VivWorkspace()
    vw.loadFromFile(opts.path)
    vw.analyze()

    retn = {}
    retn['functions'] = dumpFunctions(vw)
    retn['entrypoints'] = vw.getEntryPoints()
    retn['pointers'] = vw.getLocations(ltype=v_const.LOC_POINTER)
    retn['strings'] = vw.getLocations(ltype=v_const.LOC_STRING)
    retn['unicode'] = vw.getLocations(ltype=v_const.LOC_UNI)
    retn['imports'] = vw.getLocations(ltype=v_const.LOC_IMPORT)
    retn['exports'] = vw.getExports()
    retn['names'] = vw.getNames()
    retn['xrefs'] = vw.getXrefs()
    retn['relocations'] = vw.getRelocations()

    with open(opts.output, 'w') as fd:
        json.dump(retn, fd, indent=4)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



from PyQt5.QtWidgets import QVBoxLayout

import vtrace.qt
import vdb.qt.base

class VdbThreadsWindow(vdb.qt.base.VdbWidgetWindow):
    def __init__(self, db, dbt, parent=None):
        vdb.qt.base.VdbWidgetWindow.__init__(self, db, dbt, parent=parent)

        self.threadWidget = vtrace.qt.VQThreadsView(trace=dbt, parent=parent)

        vbox = QVBoxLayout()
        vbox.addWidget(self.threadWidget)
        self.setLayout(vbox)

        self.setWindowTitle('Threads')

    def vqLoad(self):
        '''
        the widgets in VQThreadsView already register for notifications.
        '''
        self.threadWidget.vqLoad()




# FIXME this is named wrong!

import vstruct
from vstruct.primitives import *

MAX_PATH = 260

class CLIENT_ID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.UniqueProcess = v_ptr32()
        self.UniqueThread = v_ptr32()

class EXCEPTION_RECORD(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.ExceptionCode = v_uint32()
        self.ExceptionFlags = v_uint32()
        self.ExceptionRecord = v_ptr32()
        self.ExceptionAddress = v_ptr32()
        self.NumberParameters = v_uint32()

class EXCEPTION_REGISTRATION(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.prev = v_ptr32()
        self.handler = v_ptr32()

class HEAP(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Entry = HEAP_ENTRY()
        self.Signature = v_uint32()
        self.Flags = v_uint32()
        self.ForceFlags = v_uint32()
        self.VirtualMemoryThreshold = v_uint32()
        self.SegmentReserve = v_uint32()
        self.SegmentCommit = v_uint32()
        self.DeCommitFreeBlockThreshold = v_uint32()
        self.DeCommitTotalFreeThreshold = v_uint32()
        self.TotalFreeSize = v_uint32()
        self.MaximumAllocationSize = v_uint32()
        self.ProcessHeapsListIndex = v_uint16()
        self.HeaderValidateLength = v_uint16()
        self.HeaderValidateCopy = v_ptr32()
        self.NextAvailableTagIndex = v_uint16()
        self.MaximumTagIndex = v_uint16()
        self.TagEntries = v_ptr32()
        self.UCRSegments = v_ptr32()
        self.UnusedUnCommittedRanges = v_ptr32()
        self.AlignRound = v_uint32()
        self.AlignMask = v_uint32()
        self.VirtualAllocBlocks = ListEntry()
        self.Segments = vstruct.VArray([v_uint32() for i in range(64)])
        self.u = vstruct.VArray([v_uint8() for i in range(16)])
        self.u2 = vstruct.VArray([v_uint8() for i in range(2)])
        self.AllocatorBackTraceIndex = v_uint16()
        self.NonDedicatedListLength = v_uint32()
        self.LargeBlocksIndex = v_ptr32()
        self.PseudoTagEntries = v_ptr32()
        self.FreeLists = vstruct.VArray([ListEntry() for i in range(128)])
        self.LockVariable = v_uint32()
        self.CommitRoutine = v_ptr32()
        self.FrontEndHeap = v_ptr32()
        self.FrontEndHeapLockCount = v_uint16()
        self.FrontEndHeapType = v_uint8()
        self.LastSegmentIndex = v_uint8()

class HEAP_SEGMENT(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Entry = HEAP_ENTRY()
        self.SegmentSignature = v_uint32()
        self.SegmentFlags = v_uint32()
        self.Heap = v_ptr32()
        self.LargestUncommitedRange = v_uint32()
        self.BaseAddress = v_ptr32()
        self.NumberOfPages = v_uint32()
        self.FirstEntry = v_ptr32()
        self.LastValidEntry = v_ptr32()
        self.NumberOfUnCommittedPages = v_uint32()
        self.NumberOfUnCommittedRanges = v_uint32()
        self.UncommittedRanges = v_ptr32()
        self.SegmentAllocatorBackTraceIndex = v_uint16()
        self.Reserved = v_uint16()
        self.LastEntryInSegment = v_ptr32()

class HEAP_ENTRY(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Size = v_uint16()
        self.PrevSize = v_uint16()
        self.SegmentIndex = v_uint8()
        self.Flags = v_uint8()
        self.Unused = v_uint8()
        self.TagIndex = v_uint8()

class ListEntry(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.Flink = v_ptr32()
        self.Blink = v_ptr32()

class NT_TIB(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.ExceptionList = v_ptr32()
        self.StackBase = v_ptr32()
        self.StackLimit = v_ptr32()
        self.SubSystemTib = v_ptr32()
        self.FiberData = v_ptr32()
        #x.Version = v_ptr32() # This is a union field
        self.ArbitraryUserPtr = v_ptr32()
        self.Self = v_ptr32()

class PEB(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.InheritedAddressSpace = v_uint8()
        self.ReadImageFileExecOptions = v_uint8()
        self.BeingDebugged = v_uint8()
        self.SpareBool = v_uint8()
        self.Mutant = v_ptr32()
        self.ImageBaseAddress = v_ptr32()
        self.Ldr = v_ptr32()
        self.ProcessParameters = v_ptr32()
        self.SubSystemData = v_ptr32()
        self.ProcessHeap = v_ptr32()
        self.FastPebLock = v_ptr32()
        self.FastPebLockRoutine = v_ptr32()
        self.FastPebUnlockRoutine = v_ptr32()
        self.EnvironmentUpdateCount = v_uint32()
        self.KernelCallbackTable = v_ptr32()
        self.SystemReserved = v_uint32()
        self.AtlThunkSListPtr32 = v_ptr32()
        self.FreeList = v_ptr32()
        self.TlsExpansionCounter = v_uint32()
        self.TlsBitmap = v_ptr32()
        self.TlsBitmapBits = vstruct.VArray([v_uint32() for i in range(2)])
        self.ReadOnlySharedMemoryBase = v_ptr32()
        self.ReadOnlySharedMemoryHeap = v_ptr32()
        self.ReadOnlyStaticServerData = v_ptr32()
        self.AnsiCodePageData = v_ptr32()
        self.OemCodePageData = v_ptr32()
        self.UnicodeCaseTableData = v_ptr32()
        self.NumberOfProcessors = v_uint32()
        self.NtGlobalFlag = v_uint64()
        self.CriticalSectionTimeout = v_uint64()
        self.HeapSegmentReserve = v_uint32()
        self.HeapSegmentCommit = v_uint32()
        self.HeapDeCommitTotalFreeThreshold = v_uint32()
        self.HeapDeCommitFreeBlockThreshold = v_uint32()
        self.NumberOfHeaps = v_uint32()
        self.MaximumNumberOfHeaps = v_uint32()
        self.ProcessHeaps = v_ptr32()
        self.GdiSharedHandleTable = v_ptr32()
        self.ProcessStarterHelper = v_ptr32()
        self.GdiDCAttributeList = v_uint32()
        self.LoaderLock = v_ptr32()
        self.OSMajorVersion = v_uint32()
        self.OSMinorVersion = v_uint32()
        self.OSBuildNumber = v_uint16()
        self.OSCSDVersion = v_uint16()
        self.OSPlatformId = v_uint32()
        self.ImageSubsystem = v_uint32()
        self.ImageSubsystemMajorVersion = v_uint32()
        self.ImageSubsystemMinorVersion = v_uint32()
        self.ImageProcessAffinityMask = v_uint32()
        self.GdiHandleBuffer = vstruct.VArray([v_ptr32() for i in range(34)])
        self.PostProcessInitRoutine = v_ptr32()
        self.TlsExpansionBitmap = v_ptr32()
        self.TlsExpansionBitmapBits = vstruct.VArray([v_uint32() for i in range(32)])
        self.SessionId = v_uint32()
        self.AppCompatFlags = v_uint64()
        self.AppCompatFlagsUser = v_uint64()
        self.pShimData = v_ptr32()
        self.AppCompatInfo = v_ptr32()
        self.CSDVersion = v_ptr32()
        self.UNKNOWN = v_uint32()
        self.ActivationContextData = v_ptr32()
        self.ProcessAssemblyStorageMap = v_ptr32()
        self.SystemDefaultActivationContextData = v_ptr32()
        self.SystemAssemblyStorageMap = v_ptr32()
        self.MinimumStackCommit = v_uint32()

class SEH3_SCOPETABLE(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.EnclosingLevel = v_int32()
        self.FilterFunction = v_ptr32()
        self.HandlerFunction = v_ptr32()

class SEH4_SCOPETABLE(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.GSCookieOffset = v_int32()
        self.GSCookieXOROffset = v_int32()
        self.EHCookieOffset = v_int32()
        self.EHCookieXOROffset = v_int32()
        self.EnclosingLevel = v_int32()
        self.FilterFunction = v_ptr32()
        self.HandlerFunction = v_ptr32()

class TEB(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.TIB = NT_TIB()
        self.EnvironmentPointer = v_ptr32()
        self.ClientId = CLIENT_ID()
        self.ActiveRpcHandle = v_ptr32()
        self.ThreadLocalStorage = v_ptr32()
        self.ProcessEnvironmentBlock = v_ptr32()
        self.LastErrorValue = v_uint32()
        self.CountOfOwnedCriticalSections = v_uint32()
        self.CsrClientThread = v_ptr32()
        self.Win32ThreadInfo = v_ptr32()
        self.User32Reserved = vstruct.VArray([v_uint32() for i in range(26)])
        self.UserReserved = vstruct.VArray([v_uint32() for i in range(5)])
        self.WOW32Reserved = v_ptr32()
        self.CurrentLocale = v_uint32()
        self.FpSoftwareStatusRegister = v_uint32()

class CLSID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.uuid = GUID()

class IID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.uuid = GUID()

class SYSTEMTIME(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.wYear = v_uint16()
        self.wMonth = v_uint16()
        self.wDayOfWeek = v_uint16()
        self.wDay = v_uint16()
        self.wHour = v_uint16()
        self.wMinute = v_uint16()
        self.wSecond = v_uint16()
        self.wMilliseconds = v_uint16()

    def __str__(self):
        #no timezone info
        return '%04d-%02d-%02d %02d:%02d:%02d.%d' % (self.wYear, self.wMonth, self.wDay, self.wHour, self.wMinute, self.wSecond, self.wMilliseconds)

class OSVERSIONINFOEXA(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dwOSVersionInfoSize    = v_uint32() 
        self.dwMajorVersion         = v_uint32()
        self.dwMinorVersion         = v_uint32()
        self.dwBuildNumber          = v_uint32()
        self.dwPlatformId           = v_uint32()
        self.szCSDVersion           = v_str(128)
        self.wServicePackMajor      = v_uint16()
        self.wServicePackMinor      = v_uint16()
        self.wSuiteMask             = v_uint16()
        self.wProductType           = v_uint8()
        self.wReserved              = v_uint8()

class OSVERSIONINFOEXW(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dwOSVersionInfoSize    = v_uint32() 
        self.dwMajorVersion         = v_uint32()
        self.dwMinorVersion         = v_uint32()
        self.dwBuildNumber          = v_uint32()
        self.dwPlatformId           = v_uint32()
        self.szCSDVersion           = v_wstr(128)
        self.wServicePackMajor      = v_uint16()
        self.wServicePackMinor      = v_uint16()
        self.wSuiteMask             = v_uint16()
        self.wProductType           = v_uint8()
        self.wReserved              = v_uint8()

class SERVICE_STATUS(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dwServiceType          = v_uint32()
        self.dwCurrentState         = v_uint32()
        self.dwControlsAccepted     = v_uint32()
        self.dwWin32ExitCode        = v_uint32()
        self.dwServiceSpecificExitCode = v_uint32()
        self.dwCheckPoint           = v_uint32()
        self.dwWaitHint             = v_uint32()

class KEY_EVENT_RECORD(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.bKeyDown               = v_uint32()
        self.wRepeatCount           = v_uint16()
        self.wVirtualKeyCode        = v_uint16()
        self.wVirtualScanCode       = v_uint16()
        self.unicodeChar            = v_uint16()
        self.dwControlKeyState      = v_uint32()

class WIN32_FIND_DATAW(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dwFileAttributes       = v_uint32()
        self.ftCreationTime         = v_uint64()
        self.ftLastAccessTime       = v_uint64()
        self.ftLastWriteTime        = v_uint64()
        self.nFileSizeHigh          = v_uint32()
        self.nFileSizeLow           = v_uint32()
        self.dwReserved0            = v_uint32()
        self.dwReserved1            = v_uint32()
        self.cFileName              = v_wstr(MAX_PATH)
        self.cAlternateFileName     = v_wstr(14)

class MIB_TCPROW(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dwState                = v_uint32()
        self.dwLocalAddr            = v_uint32()
        self.dwLocalPort            = v_uint32()
        self.dwRemoteAddr           = v_uint32()
        self.dwRemotePort           = v_uint32()

class MIB_UDPROW_OWNER_PID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dwLocalAddr            = v_uint32()
        self.dwLocalPort            = v_uint32()
        self.dwOwningPid            = v_uint32()


DRIVE_TYPE = v_enum()
DRIVE_TYPE.UNKNOWN              = 0
DRIVE_TYPE.NO_ROOT_DIR          = 1
DRIVE_TYPE.REMOVABLE            = 2
DRIVE_TYPE.FIXED                = 3
DRIVE_TYPE.REMOTE               = 4
DRIVE_TYPE.CDROM                = 5
DRIVE_TYPE.RAMDISK              = 6

#INPUT_RECORD_SIZE               = 0x20
#WINDOW_BUFFER_SIZE_EVENT        = 0x04
# EventType for INPUT_RECORD struct
INPUT_RECORD_EVENT_TYPE = v_enum()
INPUT_RECORD_EVENT_TYPE.KEY_EVENT      = 0x01
INPUT_RECORD_EVENT_TYPE.FOCUS_EVENT    = 0x10
INPUT_RECORD_EVENT_TYPE.MENU_EVENT     = 0x08
INPUT_RECORD_EVENT_TYPE.MOUSE_EVENT    = 0x02

FILE_OPERATION = v_enum()
FILE_OPERATION.FO_MOVE          = 0x0001
FILE_OPERATION.FO_COPY          = 0x0002
FILE_OPERATION.FO_DELETE        = 0x0003
FILE_OPERATION.FO_RENAME        = 0x0004

KEY_EVENT = v_bitmask()
KEY_EVENT.CAPSLOCK_ON           = 0x0080
KEY_EVENT.ENHANCED_KEY          = 0x0100
KEY_EVENT.LEFT_ALT_PRESSED      = 0x0002
KEY_EVENT.LEFT_CTRL_PRESSED     = 0x0008
KEY_EVENT.NUMLOCK_ON            = 0x0020
KEY_EVENT.RIGHT_ALT_PRESSED     = 0x0001
KEY_EVENT.RIGHT_CTRL_PRESSED    = 0x0004
KEY_EVENT.SCROLLLOCK_ON         = 0x0040
KEY_EVENT.SHIFT_PRESSED         = 0x0010

MIB_TCP_STATE = v_enum()
MIB_TCP_STATE.CLOSED            = 1
MIB_TCP_STATE.LISTEN            = 2
MIB_TCP_STATE.SYN_SENT          = 3
MIB_TCP_STATE.SYN_RCVD          = 4
MIB_TCP_STATE.ESTAB             = 5
MIB_TCP_STATE.FIN_WAIT1         = 6
MIB_TCP_STATE.FIN_WAIT2         = 7
MIB_TCP_STATE.CLOSE_WAIT        = 8
MIB_TCP_STATE.CLOSING           = 9
MIB_TCP_STATE.LAST_ACK          = 10
MIB_TCP_STATE.TIME_WAIT         = 11
MIB_TCP_STATE.DELETE_TCB        = 12


SERVICE_STATE = v_enum()
SERVICE_STATE.CONTINUE_PENDING = 0x00000005
SERVICE_STATE.PAUSE_PENDING    = 0x00000006
SERVICE_STATE.PAUSED           = 0x00000007
SERVICE_STATE.RUNNING          = 0x00000004
SERVICE_STATE.START_PENDING    = 0x00000002
SERVICE_STATE.STOP_PENDING     = 0x00000003
SERVICE_STATE.STOPPED          = 0x00000001

SERVICE_START_TYPE = v_enum()
SERVICE_START_TYPE.AUTO_START   = 0x00000002
SERVICE_START_TYPE.BOOT_START   = 0x00000000
SERVICE_START_TYPE.DEMAND_START = 0x00000003
SERVICE_START_TYPE.DISABLED     = 0x00000004
SERVICE_START_TYPE.SYSTEM_START = 0x00000001

SERVICE_CONTROL = v_enum()
SERVICE_CONTROL.STOP            = 0x00000001
SERVICE_CONTROL.PAUSE           = 0x00000002
SERVICE_CONTROL.CONTINUE        = 0x00000003
SERVICE_CONTROL.START           = 0x00088888  #Fake SERVICE_CONTROL value to include start here

SERVICE_TYPE = v_bitmask()
SERVICE_TYPE.FILE_SYSTEM_DRIVER  =  0x00000002
SERVICE_TYPE.KERNEL_DRIVER       =  0x00000001
SERVICE_TYPE.WIN32_OWN_PROCESS   =  0x00000010
SERVICE_TYPE.WIN32_SHARE_PROCESS =  0x00000020
SERVICE_TYPE.INTERACTIVE_PROCESS =  0x00000100

HKEY_REGISTRY = v_enum()
HKEY_REGISTRY.HKEY_CLASSES_ROOT        =  0x80000000
HKEY_REGISTRY.HKEY_CURRENT_USER        =  0x80000001
HKEY_REGISTRY.HKEY_LOCAL_MACHINE       =  0x80000002
HKEY_REGISTRY.HKEY_USERS               =  0x80000003
HKEY_REGISTRY.HKEY_PERFORMANCE_DATA    =  0x80000004
HKEY_REGISTRY.HKEY_PERFORMANCE_TEXT    =  0x80000050
HKEY_REGISTRY.HKEY_PERFORMANCE_NLSTEXT =  0x80000060
HKEY_REGISTRY.HKEY_CURRENT_CONFIG      =  0x80000005
HKEY_REGISTRY.HKEY_DYN_DATA            = 0x80000006

REG_TYPE = v_enum()
REG_TYPE.REG_NONE                        = 0
REG_TYPE.REG_SZ                          = 1
REG_TYPE.REG_EXPAND_SZ                   = 2
REG_TYPE.REG_BINARY                      = 3
REG_TYPE.REG_DWORD                       = 4
REG_TYPE.REG_DWORD_BIG_ENDIAN            = 5
REG_TYPE.REG_LINK                        = 6
REG_TYPE.REG_MULTI_SZ                    = 7
REG_TYPE.REG_RESOURCE_LIST               = 8
REG_TYPE.REG_FULL_RESOURCE_DESCRIPTOR    = 9
REG_TYPE.REG_RESOURCE_REQUIREMENTS_LIST  = 10
REG_TYPE.REG_QWORD                       = 11

FILE_ATTRIBUTE = v_bitmask()
FILE_ATTRIBUTE.READONLY             = 0x00000001  
FILE_ATTRIBUTE.HIDDEN               = 0x00000002  
FILE_ATTRIBUTE.SYSTEM               = 0x00000004  
FILE_ATTRIBUTE.DIRECTORY            = 0x00000010  
FILE_ATTRIBUTE.ARCHIVE              = 0x00000020  
FILE_ATTRIBUTE.DEVICE               = 0x00000040  
FILE_ATTRIBUTE.NORMAL               = 0x00000080  
FILE_ATTRIBUTE.TEMPORARY            = 0x00000100  
FILE_ATTRIBUTE.SPARSE_FILE          = 0x00000200  
FILE_ATTRIBUTE.REPARSE_POINT        = 0x00000400  
FILE_ATTRIBUTE.COMPRESSED           = 0x00000800  
FILE_ATTRIBUTE.OFFLINE              = 0x00001000  
FILE_ATTRIBUTE.NOT_CONTENT_INDEXED  = 0x00002000  
FILE_ATTRIBUTE.ENCRYPTED            = 0x00004000  
#INVALID_FILE_ATTRIBUTES             = 0xffffffff

WNET_RESOURCE_SCOPE = v_enum()
WNET_RESOURCE_SCOPE.RESOURCE_CONNECTED  = 0x00000001
WNET_RESOURCE_SCOPE.RESOURCE_GLOBALNET  = 0x00000002
WNET_RESOURCE_SCOPE.RESOURCE_REMEMBERED = 0x00000003
WNET_RESOURCE_SCOPE.RESOURCE_RECENT     = 0x00000004
WNET_RESOURCE_SCOPE.RESOURCE_CONTEXT    = 0x00000005


WNET_RESOURCE_TYPE = v_enum()
WNET_RESOURCE_TYPE.ANY        = 0x00000000
WNET_RESOURCE_TYPE.DISK       = 0x00000001
WNET_RESOURCE_TYPE.PRINT      = 0x00000002
WNET_RESOURCE_TYPE.RESERVED   = 0x00000008
WNET_RESOURCE_TYPE.UNKNOWN    = 0xFFFFFFFF


WNET_RESOURCEDISPLAYTYPE = v_enum()
WNET_RESOURCEDISPLAYTYPE.GENERIC        = 0x00000000
WNET_RESOURCEDISPLAYTYPE.DOMAIN         = 0x00000001
WNET_RESOURCEDISPLAYTYPE.SERVER         = 0x00000002
WNET_RESOURCEDISPLAYTYPE.SHARE          = 0x00000003
WNET_RESOURCEDISPLAYTYPE.FILE           = 0x00000004
WNET_RESOURCEDISPLAYTYPE.GROUP          = 0x00000005
WNET_RESOURCEDISPLAYTYPE.NETWORK        = 0x00000006
WNET_RESOURCEDISPLAYTYPE.ROOT           = 0x00000007
WNET_RESOURCEDISPLAYTYPE.SHAREADMIN     = 0x00000008
WNET_RESOURCEDISPLAYTYPE.DIRECTORY      = 0x00000009
WNET_RESOURCEDISPLAYTYPE.TREE           = 0x0000000A
WNET_RESOURCEDISPLAYTYPE.NDSCONTAINER   = 0x0000000B

EXIT_WINDOWS = v_bitmask()
EXIT_WINDOWS.HYBRID_SHUTDOWN  = 0x00400000 
EXIT_WINDOWS.LOGOFF           = 0 
EXIT_WINDOWS.POWEROFF         = 0x00000008 
EXIT_WINDOWS.REBOOT           = 0x00000002 
EXIT_WINDOWS.RESTARTAPPS      = 0x00000040 
EXIT_WINDOWS.SHUTDOWN         = 0x00000001 
EXIT_WINDOWS.FORCE            = 0x00000004 
EXIT_WINDOWS.FORCEIFHUNG      = 0x00000010 

WTS_CHANGE_EVENT = v_enum()
WTS_CHANGE_EVENT.WTS_CONSOLE_CONNECT             = 0x1
WTS_CHANGE_EVENT.WTS_CONSOLE_DISCONNECT          = 0x2
WTS_CHANGE_EVENT.WTS_REMOTE_CONNECT              = 0x3
WTS_CHANGE_EVENT.WTS_REMOTE_DISCONNECT           = 0x4
WTS_CHANGE_EVENT.WTS_SESSION_LOGON               = 0x5
WTS_CHANGE_EVENT.WTS_SESSION_LOGOFF              = 0x6
WTS_CHANGE_EVENT.WTS_SESSION_LOCK                = 0x7
WTS_CHANGE_EVENT.WTS_SESSION_UNLOCK              = 0x8
WTS_CHANGE_EVENT.WTS_SESSION_REMOTE_CONTROL      = 0x9
WTS_CHANGE_EVENT.WTS_SESSION_CREATE              = 0xA
WTS_CHANGE_EVENT.WTS_SESSION_TERMINATE           = 0xB

# each entry is the (vkey_name, mappable_base, shift_modified)
# if the key is non printable, None is used
# if there is no shift-modifier, None is used
VIRTUAL_KEY_CODE = v_enum()
VIRTUAL_KEY_CODE.VK_LBUTTON     = 0x01
VIRTUAL_KEY_CODE.VK_RBUTTON     = 0x02
VIRTUAL_KEY_CODE.VK_CANCEL      = 0x03
VIRTUAL_KEY_CODE.VK_MBUTTON     = 0x04
VIRTUAL_KEY_CODE.VK_XBUTTON1    = 0x05
VIRTUAL_KEY_CODE.VK_XBUTTON2    = 0x06
# 0x07: Undefined
VIRTUAL_KEY_CODE.VK_BACK        = 0x08
VIRTUAL_KEY_CODE.VK_TAB         = 0x09
# 0x0a-0x0b: Reserved
VIRTUAL_KEY_CODE.VK_CLEAR       = 0x0c
VIRTUAL_KEY_CODE.VK_RETURN      = 0x0d
# 0x0e-0x0f: Undefined
VIRTUAL_KEY_CODE.VK_SHIFT       = 0x10
VIRTUAL_KEY_CODE.VK_CONTROL     = 0x11
VIRTUAL_KEY_CODE.VK_MENU        = 0x12
VIRTUAL_KEY_CODE.VK_PAUSE       = 0x13
VIRTUAL_KEY_CODE.VK_CAPITAL     = 0x14
VIRTUAL_KEY_CODE.VK_HANGUEL     = 0x15
VIRTUAL_KEY_CODE.VK_HANGUL      = 0x15
VIRTUAL_KEY_CODE.VK_KANA        = 0x15
# 0x16: undefined
VIRTUAL_KEY_CODE.VK_JUNJA       = 0x17
VIRTUAL_KEY_CODE.VK_FINAL       = 0x18
VIRTUAL_KEY_CODE.VK_HANJA       = 0x19
VIRTUAL_KEY_CODE.VK_KANJI       = 0x19
# 0x1a: undefined
VIRTUAL_KEY_CODE.VK_ESCAPE      = 0x1b
VIRTUAL_KEY_CODE.VK_CONVERT     = 0x1c
VIRTUAL_KEY_CODE.VK_NONCONVERT  = 0x1d
VIRTUAL_KEY_CODE.VK_ACCEPT      = 0x1e
VIRTUAL_KEY_CODE.VK_MODECHANGE  = 0x1f
VIRTUAL_KEY_CODE.VK_SPACE       = 0x20
VIRTUAL_KEY_CODE.VK_PRIOR       = 0x21
VIRTUAL_KEY_CODE.VK_NEXT        = 0x22
VIRTUAL_KEY_CODE.VK_END         = 0x23
VIRTUAL_KEY_CODE.VK_HOME        = 0x24
VIRTUAL_KEY_CODE.VK_LEFT        = 0x25
VIRTUAL_KEY_CODE.VK_UP          = 0x26
VIRTUAL_KEY_CODE.VK_RIGHT       = 0x27
VIRTUAL_KEY_CODE.VK_DOWN        = 0x28
VIRTUAL_KEY_CODE.VK_SELECT      = 0x29
VIRTUAL_KEY_CODE.VK_PRINT       = 0x2a
VIRTUAL_KEY_CODE.VK_EXECUTE     = 0x2b
VIRTUAL_KEY_CODE.VK_SNAPSHOT    = 0x2c
VIRTUAL_KEY_CODE.VK_INSERT      = 0x2d
VIRTUAL_KEY_CODE.VK_DELETE      = 0x2e
VIRTUAL_KEY_CODE.VK_HELP        = 0x2f
VIRTUAL_KEY_CODE.VK_0           = 0x30
VIRTUAL_KEY_CODE.VK_1           = 0x31
VIRTUAL_KEY_CODE.VK_2           = 0x32
VIRTUAL_KEY_CODE.VK_3           = 0x33
VIRTUAL_KEY_CODE.VK_4           = 0x34
VIRTUAL_KEY_CODE.VK_5           = 0x35
VIRTUAL_KEY_CODE.VK_6           = 0x36
VIRTUAL_KEY_CODE.VK_7           = 0x37
VIRTUAL_KEY_CODE.VK_8           = 0x38
VIRTUAL_KEY_CODE.VK_9           = 0x39
# 0x3a-0x40: undefinded
VIRTUAL_KEY_CODE.VK_A           = 0x41
VIRTUAL_KEY_CODE.VK_B           = 0x42
VIRTUAL_KEY_CODE.VK_C           = 0x43
VIRTUAL_KEY_CODE.VK_D           = 0x44
VIRTUAL_KEY_CODE.VK_E           = 0x45
VIRTUAL_KEY_CODE.VK_F           = 0x46
VIRTUAL_KEY_CODE.VK_G           = 0x47
VIRTUAL_KEY_CODE.VK_H           = 0x48
VIRTUAL_KEY_CODE.VK_I           = 0x49
VIRTUAL_KEY_CODE.VK_J           = 0x4a
VIRTUAL_KEY_CODE.VK_K           = 0x4b
VIRTUAL_KEY_CODE.VK_L           = 0x4c
VIRTUAL_KEY_CODE.VK_M           = 0x4d
VIRTUAL_KEY_CODE.VK_N           = 0x4e
VIRTUAL_KEY_CODE.VK_O           = 0x4f
VIRTUAL_KEY_CODE.VK_P           = 0x50
VIRTUAL_KEY_CODE.VK_Q           = 0x51
VIRTUAL_KEY_CODE.VK_R           = 0x52
VIRTUAL_KEY_CODE.VK_S           = 0x53
VIRTUAL_KEY_CODE.VK_T           = 0x54
VIRTUAL_KEY_CODE.VK_U           = 0x55
VIRTUAL_KEY_CODE.VK_V           = 0x56
VIRTUAL_KEY_CODE.VK_W           = 0x57
VIRTUAL_KEY_CODE.VK_X           = 0x58
VIRTUAL_KEY_CODE.VK_Y           = 0x59
VIRTUAL_KEY_CODE.VK_Z           = 0x5a
VIRTUAL_KEY_CODE.VK_LWIN        = 0x5b
VIRTUAL_KEY_CODE.VK_RWIN        = 0x5c
VIRTUAL_KEY_CODE.VK_APPS        = 0x5d
# 0x5e reserved
VIRTUAL_KEY_CODE.VK_SLEEP       = 0x5f
VIRTUAL_KEY_CODE.VK_NUMPAD0     = 0x60
VIRTUAL_KEY_CODE.VK_NUMPAD1     = 0x61
VIRTUAL_KEY_CODE.VK_NUMPAD2     = 0x62
VIRTUAL_KEY_CODE.VK_NUMPAD3     = 0x63
VIRTUAL_KEY_CODE.VK_NUMPAD4     = 0x64
VIRTUAL_KEY_CODE.VK_NUMPAD5     = 0x65
VIRTUAL_KEY_CODE.VK_NUMPAD6     = 0x66
VIRTUAL_KEY_CODE.VK_NUMPAD7     = 0x67
VIRTUAL_KEY_CODE.VK_NUMPAD8     = 0x68
VIRTUAL_KEY_CODE.VK_NUMPAD9     = 0x69
VIRTUAL_KEY_CODE.VK_MULTIPLY    = 0x6a
VIRTUAL_KEY_CODE.VK_ADD         = 0x6b
VIRTUAL_KEY_CODE.VK_SEPARATOR   = 0x6c
VIRTUAL_KEY_CODE.VK_SUBTRACT    = 0x6d
VIRTUAL_KEY_CODE.VK_DECIMAL     = 0x6e
VIRTUAL_KEY_CODE.VK_DIVIDE      = 0x6f
VIRTUAL_KEY_CODE.VK_F1          = 0x70
VIRTUAL_KEY_CODE.VK_F2          = 0x71
VIRTUAL_KEY_CODE.VK_F3          = 0x72
VIRTUAL_KEY_CODE.VK_F4          = 0x73
VIRTUAL_KEY_CODE.VK_F5          = 0x74
VIRTUAL_KEY_CODE.VK_F6          = 0x75
VIRTUAL_KEY_CODE.VK_F7          = 0x76
VIRTUAL_KEY_CODE.VK_F8          = 0x77
VIRTUAL_KEY_CODE.VK_F9          = 0x78
VIRTUAL_KEY_CODE.VK_F10         = 0x79
VIRTUAL_KEY_CODE.VK_F11         = 0x7a
VIRTUAL_KEY_CODE.VK_F12         = 0x7b
VIRTUAL_KEY_CODE.VK_F13         = 0x7c
VIRTUAL_KEY_CODE.VK_F14         = 0x7d
VIRTUAL_KEY_CODE.VK_F15         = 0x7e
VIRTUAL_KEY_CODE.VK_F16         = 0x7f
VIRTUAL_KEY_CODE.VK_F17         = 0x80
VIRTUAL_KEY_CODE.VK_F18         = 0x81
VIRTUAL_KEY_CODE.VK_F19         = 0x82
VIRTUAL_KEY_CODE.VK_F20         = 0x83
VIRTUAL_KEY_CODE.VK_F21         = 0x84
VIRTUAL_KEY_CODE.VK_F22         = 0x85
VIRTUAL_KEY_CODE.VK_F23         = 0x86
VIRTUAL_KEY_CODE.VK_F24         = 0x87
# 0x88-0x8f: unassigned
VIRTUAL_KEY_CODE.VK_NUMLOCK     = 0x90
VIRTUAL_KEY_CODE.VK_SCROLL      = 0x91
# 0x92-0x96: OEM specific
# 0x97-0x9f: unassigned
VIRTUAL_KEY_CODE.VK_LSHIFT      = 0xa0
VIRTUAL_KEY_CODE.VK_RSHIFT      = 0xa1
VIRTUAL_KEY_CODE.VK_LCONTROL    = 0xa2
VIRTUAL_KEY_CODE.VK_RCONTROL    = 0xa3
VIRTUAL_KEY_CODE.VK_LMENU       = 0xa4
VIRTUAL_KEY_CODE.VK_RMENU       = 0xa5
VIRTUAL_KEY_CODE.VK_BROWSER_BACK = 0xa6
VIRTUAL_KEY_CODE.VK_BROWSER_FORWARD = 0xa7
VIRTUAL_KEY_CODE.VK_BROWSER_REFRESH = 0xa8
VIRTUAL_KEY_CODE.VK_BROWSER_STOP = 0xa9
VIRTUAL_KEY_CODE.VK_BROWSER_SEARCH = 0xaa
VIRTUAL_KEY_CODE.VK_BROWSER_FAVORITES = 0xab
VIRTUAL_KEY_CODE.VK_BROWSER_HOME = 0xac
VIRTUAL_KEY_CODE.VK_VOLUME_MUTE = 0xad
VIRTUAL_KEY_CODE.VK_VOLUME_DOWN = 0xae
VIRTUAL_KEY_CODE.VK_VOLUME_UP   = 0xaf
VIRTUAL_KEY_CODE.VK_MEDIA_NEXT_TRACK = 0xb0
VIRTUAL_KEY_CODE.VK_MEDIA_PREV_TRACK = 0xb1
VIRTUAL_KEY_CODE.VK_MEDIA_STOP  = 0xb2
VIRTUAL_KEY_CODE.VK_MEDIA_PLAY_PAUSE = 0xb3
VIRTUAL_KEY_CODE.VK_LAUNCH_MAIL = 0xb4
VIRTUAL_KEY_CODE.VK_LAUNCH_MEDIA_SELECT = 0xb5
VIRTUAL_KEY_CODE.VK_LAUNCH_APP1 = 0xb6
VIRTUAL_KEY_CODE.VK_LAUNCH_APP2 = 0xb7
# 0xb8-0xb9: reserved
VIRTUAL_KEY_CODE.VK_OEM_1       = 0xba
VIRTUAL_KEY_CODE.VK_OEM_PLUS    = 0xbb
VIRTUAL_KEY_CODE.VK_OEM_COMMA   = 0xbc
VIRTUAL_KEY_CODE.VK_OEM_MINUS   = 0xbd
VIRTUAL_KEY_CODE.VK_OEM_PERIOD  = 0xbe
VIRTUAL_KEY_CODE.VK_OEM_2       = 0xbf
VIRTUAL_KEY_CODE.VK_OEM_3       = 0xc0
# 0xc1-0xd7: reserved
# 0xd8-0xda: unassigned
VIRTUAL_KEY_CODE.VK_OEM_4       = 0xdb
VIRTUAL_KEY_CODE.VK_OEM_5       = 0xdc
VIRTUAL_KEY_CODE.VK_OEM_6       = 0xdd
VIRTUAL_KEY_CODE.VK_OEM_7       = 0xde
VIRTUAL_KEY_CODE.VK_OEM_8       = 0xdf
# 0xe0: reserved
# 0xe1: OEM specific
VIRTUAL_KEY_CODE.VK_OEM_102     = 0xe2
# 0xe3-0xe4: OEM specific
VIRTUAL_KEY_CODE.VK_PROCESSKEY  = 0xe5
# 0xe6: OEM specific
VIRTUAL_KEY_CODE.VK_PACKET      = 0xe7
# 0xe8: unassigned
# 0xe9-0xf5: OEM specific
VIRTUAL_KEY_CODE.VK_ATTN        = 0xf6
VIRTUAL_KEY_CODE.VK_CRSEL       = 0xf7
VIRTUAL_KEY_CODE.VK_EXSEL       = 0xf8
VIRTUAL_KEY_CODE.VK_EREOF       = 0xf9
VIRTUAL_KEY_CODE.VK_PLAY        = 0xfa
VIRTUAL_KEY_CODE.VK_ZOOM        = 0xfb
VIRTUAL_KEY_CODE.VK_NONAME      = 0xfc
VIRTUAL_KEY_CODE.VK_PA1         = 0xfd
VIRTUAL_KEY_CODE.VK_OEM_CLEAR   = 0xfe

#printable keys for US keyboard 
# tuples of (base, shift-modified)
VIRTUAL_KEY_CODE_US_MAPPING = {
    VIRTUAL_KEY_CODE.VK_RETURN  : ('\n', '\n'),
    VIRTUAL_KEY_CODE.VK_SPACE   : (' ', ' '),
    VIRTUAL_KEY_CODE.VK_0       : ('0', ')'),
    VIRTUAL_KEY_CODE.VK_1       : ('1', '!'),
    VIRTUAL_KEY_CODE.VK_2       : ('2', '@'),
    VIRTUAL_KEY_CODE.VK_3       : ('3', '#'),
    VIRTUAL_KEY_CODE.VK_4       : ('4', '$'),
    VIRTUAL_KEY_CODE.VK_5       : ('5', '%'),
    VIRTUAL_KEY_CODE.VK_6       : ('6', '^'),
    VIRTUAL_KEY_CODE.VK_7       : ('7', '&'),
    VIRTUAL_KEY_CODE.VK_8       : ('8', '*'),
    VIRTUAL_KEY_CODE.VK_9       : ('9', '('),
    VIRTUAL_KEY_CODE.VK_OEM_1   : (';', ':'),
    VIRTUAL_KEY_CODE.VK_OEM_PLUS    : ('=', '+'),
    VIRTUAL_KEY_CODE.VK_OEM_COMMA   : (',', '<'),
    VIRTUAL_KEY_CODE.VK_OEM_MINUS   : ('-', '_'),
    VIRTUAL_KEY_CODE.VK_OEM_PERIOD  : ('.', '>'),
    VIRTUAL_KEY_CODE.VK_OEM_2   : ('/', '?'),
    VIRTUAL_KEY_CODE.VK_OEM_3   : ('`', '~'),
    VIRTUAL_KEY_CODE.VK_OEM_4   : ('[', '{'),
    VIRTUAL_KEY_CODE.VK_OEM_5   : ('\\', '|'),
    VIRTUAL_KEY_CODE.VK_OEM_6   : (']', '}'),
    VIRTUAL_KEY_CODE.VK_OEM_7   : ("'", '"'),
    VIRTUAL_KEY_CODE.VK_OEM_8   : ('!', None), #section symbol &sect;
    VIRTUAL_KEY_CODE.VK_A       : ('a', 'A'),
    VIRTUAL_KEY_CODE.VK_B       : ('b', 'B'),
    VIRTUAL_KEY_CODE.VK_C       : ('c', 'C'),
    VIRTUAL_KEY_CODE.VK_D       : ('d', 'D'),
    VIRTUAL_KEY_CODE.VK_E       : ('e', 'E'),
    VIRTUAL_KEY_CODE.VK_F       : ('f', 'F'),
    VIRTUAL_KEY_CODE.VK_G       : ('g', 'G'),
    VIRTUAL_KEY_CODE.VK_H       : ('h', 'H'),
    VIRTUAL_KEY_CODE.VK_I       : ('i', 'I'),
    VIRTUAL_KEY_CODE.VK_J       : ('j', 'J'),
    VIRTUAL_KEY_CODE.VK_K       : ('k', 'K'),
    VIRTUAL_KEY_CODE.VK_L       : ('l', 'L'),
    VIRTUAL_KEY_CODE.VK_M       : ('m', 'M'),
    VIRTUAL_KEY_CODE.VK_N       : ('n', 'N'),
    VIRTUAL_KEY_CODE.VK_O       : ('o', 'O'),
    VIRTUAL_KEY_CODE.VK_P       : ('p', 'P'),
    VIRTUAL_KEY_CODE.VK_Q       : ('q', 'Q'),
    VIRTUAL_KEY_CODE.VK_R       : ('r', 'R'),
    VIRTUAL_KEY_CODE.VK_S       : ('s', 'S'),
    VIRTUAL_KEY_CODE.VK_T       : ('t', 'T'),
    VIRTUAL_KEY_CODE.VK_U       : ('u', 'U'),
    VIRTUAL_KEY_CODE.VK_V       : ('v', 'V'),
    VIRTUAL_KEY_CODE.VK_W       : ('w', 'W'),
    VIRTUAL_KEY_CODE.VK_X       : ('x', 'X'),
    VIRTUAL_KEY_CODE.VK_Y       : ('y', 'Y'),
    VIRTUAL_KEY_CODE.VK_Z       : ('z', 'Z'),
}




'''
'''
import inspect

import vstruct
import vivisect.symboliks

from vivisect.symboliks.common import *

class CallingConventionProxy:

    def __init__(self, cconv, argv, funcsym):
        self.argv = argv # viv style (type,name) tuple list
        self.cconv = cconv
        self.funcsym = funcsym

    def __call__(self, emu):

        # Get and update the symbolik args
        args = self.cconv.getSymbolikArgs(emu, self.argv)
        args = [ arg.update(emu) for arg in args ]

        # If callFunction returns something, snap it back in.
        # Otherwise, snap in a Call symbol.
        ret = self.callFunction(emu, *args)
        if ret is None:
            ret = Call(self.funcsym, emu.__width__, args)

        # Set the return value into the symbolik state
        self.cconv.setSymbolikReturn(emu, ret, self.argv)

    def getSymbolikArgs(self, emu):
        return self.cconv.getSymbolikArgs(emu, self.argv)

    def callFunction(emu, *args):
        # Each calling convention proxy must implement this to do
        # the actual call hook...
        return None


class ImportCallProxy(CallingConventionProxy):
    '''
    A calling convention proxy allows the definition of
    a pythonic function which may then be called by an emulator
    during symbolik effect processing.
    '''

    def __init__(self, func, cconv):

        # Do crazy introspection shit to make calling convention
        # map function args to names / vstruct types.
        aspec = inspect.getargspec(func)
        argn = aspec.args[1:]
        argt = aspec.defaults

        argv = [ (argt[i],argn[i]) for i in range(len(argn)) ]

        modlast = func.__module__.split('.')[-1]
        funcsym = Var('%s.%s' % (modlast, func.__name__))

        CallingConventionProxy.__init__(self, cconv, argv, funcsym)

        self.func = func

    def callFunction(self, emu, *args):
        return self.func(emu, *args)



"""
A function analysis module to detect and properly mark
microsoft hotpatch pads...
"""


def analyzeFunction(vw, funcva):

    offset, bytez = vw.getByteDef(funcva)
    ob = bytez[offset-1]
    if ob not in [0x90, 0xcc]:
        return

    count = 1
    newb = bytez[offset-count]
    while newb == ob:
        count += 1
        newb = bytez[offset-count]

    count -= 1

    va = funcva - count
    if count >= 5 and not vw.isLocation(va):
        vw.makePad(va, count)



import time
import logging
import collections

import cobra

logger = logging.getLogger(__name__)


class PoolMethod(cobra.CobraMethod):
    def __init__(self, pool, methname):
        cobra.CobraMethod.__init__(self, None, methname)
        self._cobra_pool = pool

    def __call__(self, *args, **kwargs):
        proxy = self._cobra_pool.cobraPoolGetProxy(rotate=self._cobra_pool.loadbalance)
        while True:
            try:
                return getattr(proxy, self.methname)(*args, **kwargs)
            except cobra.CobraException as e:
                logger.warning('CobraPool Method Failure: %s %s', proxy._cobra_uri, e)
                proxy = self._cobra_pool.cobraPoolGetProxy(rotate=True)


class CobraPool:
    '''
    A transparent "pool" of cobra proxies with failure,
    load balancing, and data agregation capabilties.

    NOTE: by default, all CobraExceptions are "failures".
          ( this means auth fail etc may be masked! )
    '''
    def __init__(self, uris, loadbalance=False):
        self.uris = collections.deque([str(u) for u in uris])
        self.faildelay = 1
        self.proxycache = {}
        self.loadbalance = loadbalance

    def __getattr__(self, name):
        proxy = self.cobraPoolGetProxy()
        if proxy._cobra_methods.get(name, False):
            poolmeth = PoolMethod(self, name)
            setattr(self, name, poolmeth)
            return poolmeth
        raise AttributeError('no such method: %s' % name)

    def cobraPoolAddUri(self, uri):
        '''
        Add a URI to this CobraPool.
        '''
        self.uris.append(uri)

    def cobraPoolGetProxy(self, rotate=False):
        if rotate:
            self.uris.rotate(1)

        while True:
            try:

                uri = self.uris[0]
                proxy = self.proxycache.get(uri)
                if proxy is None:
                    proxy = cobra.CobraProxy(uri)
                    self.proxycache[uri] = proxy
                return proxy
            except Exception as e:
                self.uris.rotate(1)
                logger.warning('CobraPool Proxy Failure: %s %s', uri, e)
                time.sleep(self.faildelay)  # FIXME track per?



import math
import operator

import envi
import envi.bits as e_bits
import envi.archs.i386 as e_i386
import envi.archs.i386.opconst as i386_opconst

import vivisect.symboliks.analysis as vsym_analysis
import vivisect.symboliks.callconv as vsym_callconv
import vivisect.symboliks.translator as vsym_trans

from vivisect.const import *
from vivisect.symboliks.common import *

mulpairs = {
    2: ('ax', 'dx'),
    4: ('eax', 'edx'),
    8: ('rax', 'rdx')
}

def getSegmentSymbol(op):
    if op.prefixes & e_i386.PREFIX_CS:
        return Var('cs', 4)
    if op.prefixes & e_i386.PREFIX_SS:
        return Var('ss', 4)
    if op.prefixes & e_i386.PREFIX_DS:
        return Var('ds', 4)
    if op.prefixes & e_i386.PREFIX_ES:
        return Var('es', 4)
    if op.prefixes & e_i386.PREFIX_FS:
        return Var('fs', 4)
    if op.prefixes & e_i386.PREFIX_GS:
        return Var('gs', 4)

class IntelSymbolikTranslator(vsym_trans.SymbolikTranslator):
    '''
    Common parent for i386 and x64.  Make sure you define:
    __arch__, __ip__, __sp__, __srcp__, __destp__

    Symbolik representations that are 'agnostic' to 32 or 64-bit belong here.
    '''
    def __init__(self, vw):
        vsym_trans.SymbolikTranslator.__init__(self, vw)
        self._arch = self.__arch__()
        self._psize = self._arch.getPointerSize()
        self._reg_ctx = self._arch.archGetRegCtx()
        self._sz_masks = {}
        for i in (0, 1, 2, 4, 8, 16, 32):
            self._sz_masks[i] = Const((2 ** (i * 8)) - 1, self._psize)

    def getRegObj(self, regidx):
        ridx = regidx & 0xffff
        rname = self._reg_ctx.getRegisterName(ridx)
        rbitwidth = self._reg_ctx.getRegisterWidth(ridx)
        val = Var(rname, rbitwidth >> 3)

        # Translate to meta if needed...
        if ridx != regidx:
            # we cannot call _xlateToMetaReg since we'd pass in a symbolik
            # object that would trigger an and operation; the code in envi
            # obviously is NOT symboliks aware (2nd op in & operation is NOT
            # a symbolik); so we do it manually here since we are symbolik
            # aware.
            rreg, lshift, mask = self._reg_ctx.getMetaRegInfo(regidx)
            metawidth = e_bits.u_maxes.index(mask)
            if lshift != 0:
                val = o_rshift(val, Const(lshift, metawidth), metawidth)

            # We must be explicit about the width!
            val = o_and(val, Const(mask, metawidth), metawidth)

        return val

    def setRegObj(self, regidx, obj):
        ridx = regidx & 0xffff
        rname = self._reg_ctx.getRegisterName(ridx)
        rbitwidth = self._reg_ctx.getRegisterWidth(ridx)
        val = Var(rname, rbitwidth >> 3)

        # Translate to native if needed...
        if ridx != regidx:
            # we cannot call _xlateToNativReg since we'd pass in a symbolik
            # object that would trigger an or operation; the code in envi
            # obviously is NOT symboliks aware (2nd op in | operation is NOT
            # a symbolik); so we do it manually here since we are symbolik
            # aware.
            # obj = self._reg_ctx._xlateToNativeReg(regidx, obj)
            basemask = (2**rbitwidth) - 1
            rreg, lshift, mask = self._reg_ctx.getMetaRegInfo(regidx)
            # cut hole in mask
            finalmask = basemask ^ (mask << lshift)
            if lshift != 0:
                obj <<= Const(lshift, rbitwidth >> 3)

            obj = obj | (val & Const(finalmask, rbitwidth >> 3))

        self.effSetVariable(rname, obj)

    def setRegByName(self, name, obj):
        '''
        Sets the symbolik variable representing the register by name.

        This is meta register aware, so if you call setRegByName('al', Var("thingy", 1))
        it will set the eax variable (on i386) or the rax variable (on amd64) by only setting
        the lower bits (for al/ax). The upper bits/symbolik effects will remain unchained.
        In the case of setRegByName('eax', obj) on either i386 or amd64, it will clobber the
        entire register.
        '''
        ridx = self._reg_ctx.getRegisterIndex(name)
        width = self._reg_ctx.getRegisterWidth(ridx)
        if width >= 32:
            ridx = ridx & 0xffff
        self.setRegObj(ridx, obj)

    def getRegByName(self, name):
        regidx = self._reg_ctx.getRegisterIndex(name)
        return self.getRegObj(regidx)

    def getOperAddrObj(self, op, idx):
        oper = op.opers[idx]
        seg = getSegmentSymbol(op)

        if isinstance(oper, e_i386.i386RegMemOper):
            reg = Var(self._reg_ctx.getRegisterName(oper.reg), oper.tsize)
            if oper.disp == 0:
                base = reg
            if oper.disp > 0:
                base = o_add(reg, Const(oper.disp, self._psize), self._psize)
            if oper.disp < 0:
                base = o_sub(reg, Const(abs(oper.disp), self._psize), self._psize)

            if seg:
                return o_add(seg, base, self._psize)

            return base

        elif isinstance(oper, e_i386.i386SibOper):

            base = None
            if oper.imm is not None:
                base = Const(oper.imm, self._psize)

            if oper.reg is not None:
                robj = self.getRegObj(oper.reg)
                if base is None:
                    base = robj
                else:
                    base = o_add(base, robj, self._psize)

            # Base is set by here for sure!
            if oper.index is not None:
                robj = self.getRegObj(oper.index)
                base = o_add(base, o_mul(robj, Const(oper.scale, self._psize), self._psize), self._psize)

            if oper.disp is not None:
                if oper.disp > 0:
                    base = o_add(base, Const(oper.disp, self._psize), self._psize)
                else:
                    base = o_sub(base, Const(abs(oper.disp), self._psize), self._psize)

            if seg:
                return o_add(seg, base, self._psize)

            return base

        elif isinstance(oper, e_i386.i386ImmMemOper):
            # FIXME plumb more segment awareness!
            if seg:
                return o_add(seg, Const(oper.imm, self._psize), self._psize)

            return Const(oper.imm, self._psize)

    def getOperObj(self, op, idx):
        '''
        Translate the specified operand to a symbol compatible with
        the symbolic translation system.
        '''
        oper = op.opers[idx]
        if oper.isReg():
            return self.getRegObj(oper.reg)

        elif oper.isDeref():
            addrsym = self.getOperAddrObj(op, idx)
            return self.effReadMemory(addrsym, Const(oper.tsize, self._psize))

        elif oper.isImmed():
            ret = oper.getOperValue(op, self)
            return Const(ret, self._psize)

        raise Exception('Unknown operand class: %s' % oper.__class__.__name__)

    def setOperObj(self, op, idx, obj):
        '''
        Set the operand to the new symbolic object.
        '''
        oper = op.opers[idx]
        if oper.isReg():
            self.setRegObj(oper.reg, obj)
            return

        if oper.isDeref():
            addrsym = self.getOperAddrObj(op, idx)
            return self.effWriteMemory(addrsym, Const(oper.tsize, self._psize), obj)

        raise Exception('setOperObj failed')

    def _generate_parity(self, obj):
        '''
        So normally, the parity of an integer is the XOR of all of its bits. To set the
        parity bit, we just bitflip the xor of all the bits.
        '''
        valu = obj
        iters = int(math.log(self._psize * 8, 2))
        shift = 2 ** (iters - 1)
        for i in range(iters):
            valu = (valu >> Const(shift, self._psize)) ^ (valu & self._sz_masks[shift >> 1])
            shift >>= 1

        # intel's parity bit is set to 1 if the number of bits if even, and the above calculates to
        # 1 if the number of bits is odd
        return eq(valu & Const(1, self._psize), Const(0, self._psize))

    def __get_dest_maxes(self, op):
        tsize = op.opers[0].tsize
        smax = e_bits.s_maxes[tsize]
        umax = e_bits.u_maxes[tsize]
        return smax, umax

    def invert(self, oper):
        width = oper.getWidth()
        inv = oper ^ Const(e_bits.u_maxes[width], width)
        return inv

    def i_adc(self, op, isDX=False):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        dsize = op.opers[0].tsize

        add = v1 + v2 + Var('eflags_cf', 1)

        f = gt(add, Const(e_bits.u_maxes[dsize>>1], dsize))
        if isDX:
            self.effSetVariable('eflags_cf', f)
        else:
            self.effSetVariable('eflags_gt', gt(v1, v2))
            self.effSetVariable('eflags_lt', lt(v1, v2))
            self.effSetVariable('eflags_sf', lt(add, Const(0, self._psize)))
            self.effSetVariable('eflags_eq', eq(add, Const(0, self._psize)))
            self.effSetVariable('eflags_pf', self._generate_parity(add))
            self.effSetVariable('eflags_of', f)
            self.effSetVariable('eflags_cf', f)

        self.setOperObj(op, 0, add)

    def i_adcx(self, op):
        self.i_adc(op, isDX=True)

    def i_add(self, op, isDOX=False):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)

        dsize = op.opers[0].tsize
        # dmax = e_bits.s_maxes[dsize]
        # ssize = op.opers[1].tsize

        smax, umax = self.__get_dest_maxes(op)

        add = o_add(v1, v2, dsize)
        if isDOX:
            of = Var('eflags_of', self._psize)
            add = o_add(add, of, dsize)
        self.setOperObj(op, 0, add)

        self.effSetVariable('eflags_gt', gt(v1, v2))
        self.effSetVariable('eflags_lt', lt(v1, v2))
        self.effSetVariable('eflags_sf', lt(add, Const(0, self._psize)))
        self.effSetVariable('eflags_eq', eq(add, Const(0, self._psize)))
        self.effSetVariable('eflags_pf', self._generate_parity(add))
        f = gt(add, Const(e_bits.s_maxes[dsize>>1], dsize))
        self.effSetVariable('eflags_of', f)

    def i_adox(self, op):
        self.i_add(op, isDOX=True)

    def i_addsd(self, op, off=0):
        dsize = op.opers[0].tsize
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off+1)
        add = o_add(v1, v2, dsize)
        self.setOperObj(op, 0, add)

    i_addss = i_addsd

    def i_vaddsd(self, off=0):
        self.i_addsd(op, off=1)

    def i_vaddss(self, op):
        self.i_addss(op, off=1)

    def i_and(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)

        obj = o_and(v1, v2, v1.getWidth())

        self.effSetVariable('eflags_gt', gt(v1, v2))
        self.effSetVariable('eflags_lt', lt(v1, v2))
        self.effSetVariable('eflags_sf', lt(obj, Const(0, self._psize)))  # v1 & v2 < 0
        self.effSetVariable('eflags_eq', eq(obj, Const(0, self._psize)))  # v1 & v2 == 0
        self.effSetVariable('eflags_of', Const(0, self._psize))
        self.effSetVariable('eflags_cf', Const(0, self._psize))
        self.effSetVariable('eflags_pf', self._generate_parity(obj))

        self.setOperObj(op, 0, obj)

    def i_bt(self, op):
        oper = self.getOperObj(op, 0)
        bit = self.getOperObj(op, 1)
        cf = (oper >> bit) & Const(1, 1)
        self.effSetVariable('eflags_cf', cf)

    def i_btc(self, op):
        oper = self.getOperObj(op, 0)
        opersize = oper.getWidth()
        bit = self.getOperObj(op, 1)

        val = oper ^ (Const(1, opersize) << bit)
        cf = (oper >> bit) & Const(1, 1)

        self.effSetVariable('eflags_cf', cf)
        self.setOperObj(op, 0, val)

    def i_bts(self, op):
        oper = self.getOperObj(op, 0)
        opersize = oper.getWidth()
        bit = self.getOperObj(op, 1)
        if bit.isDiscrete():
            mask = Const(1 << bit.solve(), opersize)
        else:
            mask = Const(1, self._psize) << bit
        val = oper | mask
        bitinfo = (oper >> bit) & Const(1, opersize)

        self.effSetVariable('eflags_cf', bitinfo)
        self.setOperObj(op, 0, val)

    def i_btr(self, op):
        oper = self.getOperObj(op, 0)
        bit = self.getOperObj(op, 1)
        opersize = oper.getWidth()
        if bit.isDiscrete():
            mask = Const(-1 ^ (1 << bit.solve()), opersize)
        else:
            mask = Const(-1, opersize) ^ (Const(1, opersize) << bit)

        val = oper & mask
        bitinfo = (oper >> bit) & Const(1, opersize)

        self.setOperObj(op, 0, val)
        self.effSetVariable('eflags_cf', bitinfo)

    def i_call(self, op):
        # For now, calling means finding which of our symbols go in
        # and logging what comes out.
        targ = self.getOperObj(op, 0)
        self.effFofX(targ)

    def i_cld(self, op):
        self.effSetVariable('eflags_df', Const(0, self._psize))

    def i_cmp(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        res = o_sub(v1, v2, v1.getWidth())
        self.effSetVariable('eflags_cf', gt(v2, v1))  #
        self.effSetVariable('eflags_gt', gt(v1, v2))  # v1 - v2 > 0 :: v1 > v2
        self.effSetVariable('eflags_lt', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_sf', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_eq', eq(v1, v2))  # v1 - v2 == 0 :: v1 == v2

    def i_dec(self, op):
        v1 = self.getOperObj(op, 0)
        zero = Const(0, self._psize)
        one = Const(1, self._psize)

        sub = o_sub(v1, one, v1.getWidth())

        self.effSetVariable('eflags_gt', gt(sub, zero))  # v1 - 1 > 0 :: v1 > 1
        self.effSetVariable('eflags_lt', lt(sub, zero))  # v1 - 1 < 0 :: v1 < 1
        self.effSetVariable('eflags_sf', lt(sub, zero))  # v1 - 1 < 0 :: v1 < 1
        self.effSetVariable('eflags_eq', eq(sub, zero))  # v1 - 1 == 0 :: v1 == 1

        self.setOperObj(op, 0, sub)

    def _div(self, op, isInvalid=None):
        oper = op.opers[0]
        dsize = oper.tsize
        divbase = self.getOperObj(op, 0)
        if isInvalid is None:
            limit = (2 ** (oper.tsize * 8)) - 1
            isInvalid = lambda val: val > limit

        if dsize == 1:
            ax = self.getRegByName('ax')
            quot = ax / divbase
            rem = ax % divbase
            if quot.isDiscrete() and isInvalid(quot):
                raise envi.DivideError('i386 #DE')
            self.setRegByName('ax', (rem << 8) | quot)

        elif dsize == 2:
            ax = self.getRegByName('ax')
            dx = self.getRegByName('dx')
            tot = (dx << Const(16, self._psize)) | ax
            quot = tot / divbase
            rem = tot % divbase
            if quot.isDiscrete() and isInvalid(quot):
                raise envi.DivideError('i386 #DE')
            self.setRegByName('ax', quot)
            self.setRegByName('dx', rem)

        elif dsize == 4:
            eax = self.getRegByName('eax')
            edx = self.getRegByName('edx')
            tot = (edx << Const(32, self._psize)) | eax
            quot = tot / divbase
            rem = tot % divbase
            if quot.isDiscrete() and isInvalid(quot):
                raise envi.DivideError('i386 #DE')
            self.setRegByName('eax', quot)
            self.setRegByName('edx', rem)
            # FIXME maybe we need a "check exception" effect?

        else:
            raise envi.UnsupportedInstruction(self, op)

    def i_div(self, op):
        return self._div(op)

    def i_idiv(self, op):
        tsize = op.opers[0].tsize
        limit = ((-2 ** (tsize * 8 - 1)), 2 ** (tsize * 8 - 1) - 1)
        return self._div(op, isInvalid=lambda val: val < limit[0] or val > limit[1])

    def i_divsd(self, op, off=0):
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off+1)
        if v2 == 0:
            raise Exception('#DE, divide error')
        res = v1 / v2
        self.setOperObj(op, 0, res)
    i_divss = i_divsd

    def i_vdivsd(self, op):
        self.i_divsd(op, off=1)
    i_vdivss = i_vdivsd

    def i_hlt(self, op):
        # Nothing to do symbolically....
        pass

    def i_imul(self, op):
        ocount = len(op.opers)
        dsize = op.opers[0].tsize
        if ocount == 1:
            mult = self.getOperObj(op, 0)
            if dsize == 1:
                reg = self.getRegByName('al')
                res = reg * mult
                dreg = res >> Const(8, dsize)
                self.setRegByName('ax', res)
            # TODO: collapse
            elif dsize in mulpairs:
                a, d = mulpairs[dsize]
                reg = self.getRegByName(a)
                res = reg * mult
                self.setRegByName(a, res & self._sz_masks[dsize])
                dreg = res >> Const((dsize * 8), dsize)
                self.setRegByName(d, dreg)
            else:
                raise Exception('Improper size of %d for i_imul', dsize)

        elif ocount == 2:
            src1 = self.getOperObj(op, 0)
            src2 = self.getOperObj(op, 1)

            res = src1 * src2
            self.setOperObj(op, 0, res)

        elif ocount == 3:
            src1 = self.getOperObj(op, 1)
            dsize = src1.getWidth()

            src2 = self.getOperObj(op, 2)
            # to deal with 6B vs 69 versions of imul
            if src1.getWidth() > src2.getWidth():
                src2 = o_sextend(src2, Const(dsize, self._psize))

            res = src1 * src2
            self.setOperObj(op, 0, res)
        else:
            raise Exception("WTFO?  i_imul with no opers")

        chop = o_sextend(self.getOperObj(op, 0), Const(2 * op.opers[0].tsize, dsize))
        f = eq(res, chop)
        self.effSetVariable('eflags_cf', f)
        self.effSetVariable('eflags_of', f)

    def i_cdq(self, op):
        obj = self.getRegByName('ax')
        v0 = o_sextend(obj, Const(self._psize, self._psize))

        v1 = o_sextend(obj, Const(2 * self._psize, self._psize))
        v2 = v1 >> Const(8 * self._psize, self._psize)

        self.setRegByName('eax', v0)
        self.setRegByName('edx', v2)

    def i_mul(self, op):
        mult = self.getOperObj(op, 0)
        dsize = op.opers[0].tsize
        bitsize = Const(dsize * 8, dsize)
        if dsize == 1:
            reg = self.getRegByName('al')
            res = reg * mult
            dreg = res >> Const(8, dsize)
            self.setRegByName('ax', res)
        # TODO: collapse
        elif dsize in mulpairs:
            a, d = mulpairs[dsize]
            reg = self.getRegByName(a)
            res = reg * mult
            self.setRegByName(a, res & self._sz_masks[dsize])
            dreg = res >> bitsize
            self.setRegByName(d, dreg)
        else:
            raise Exception('Improper size of %d for i_mul', dsize)

        f = ne(dreg, Const(0, dsize))
        self.effSetVariable('eflags_cf', f)
        self.effSetVariable('eflags_of', f)

    def i_mulx(self, op):
        src = self.getOperObj(op, 2)
        dsize = src.getWidth()
        if dsize == 4:
            dreg = self.getRegByName('edx')
        elif dsize == 8:
            dreg = self.getRegByName('rdx')
        else:
            raise Exception('mulx has invalid size of %d' % dsize)

        # sooo, technically the size here is wrong so we need to promote it?
        res = dreg * src

        self.setOperObj(op, 0, res & self._sz_masks[dsize])
        self.setOperObj(op, 1, res >> Const(8 * dsize, self._psize))

    def i_mulsd(self, op, off=0):
        '''
        Also doesn't set flags?
        '''
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off + 1)
        ocount = len(op.opers)
        res = o_mul(v1, v2, v1.getWidth())
        self.setOperObj(op, 0, res)
    i_mulss = i_mulsd

    def i_vmulsd(self, op):
        self.i_mulsd(op, off=1)
    i_vmulss = i_vmulsd

    def i_inc(self, op):
        v1 = self.getOperObj(op, 0)
        dsize = op.opers[0].tsize
        obj = o_add(v1, Const(1, self._psize), v1.getWidth())
        zero = Const(0, self._psize)

        self.effSetVariable('eflags_gt', gt(obj, zero))
        self.effSetVariable('eflags_lt', lt(obj, zero))
        self.effSetVariable('eflags_sf', lt(obj, zero))
        self.effSetVariable('eflags_eq', eq(obj, zero))

        f = gt(obj, Const(e_bits.s_maxes[dsize>>1], dsize))
        self.effSetVariable('eflags_of', f)

        self.effSetVariable('eflags_pf', self._generate_parity(obj))
        self.setOperObj(op, 0, obj)

    def i_int3(self, op):
        # FIXME another "test for exception" effect vote!
        pass

    # We include all the possible Jcc names just in case somebody
    # gets hinkey with the disassembler.
    def i_ja(self, op):
        return self._cond_jmp(op, Var('eflags_gt', self._psize))

    def i_jae(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_lt', self._psize)))

    def i_jb(self, op):
        return self._cond_jmp(op, Var('eflags_lt', self._psize))

    def i_jbe(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_gt', self._psize)))

    def i_jc(self, op):
        return self._cond_jmp(op, Var('eflags_cf', self._psize))

    def i_je(self, op):
        return self._cond_jmp(op, Var('eflags_eq', self._psize))

    def _cond_jmp(self, op, cond):
        # Construct the tuple for the conditional jump
        return ((self.getOperObj(op, 0), cond ),
                (Const(op.va + len(op), self._psize), cnot(cond)))

    def i_jg(self, op):
        return self._cond_jmp(op, Var('eflags_gt', self._psize))

    def i_jge(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_lt', self._psize)))

    def i_jl(self, op):
        return self._cond_jmp(op, Var('eflags_lt', self._psize))

    def i_jle(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_gt', self._psize)))

    def i_jo(self, op):
        return self._cond_jmp(op, Var('eflags_of', self._psize))

    def i_jno(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_of', self._psize)))

    i_jna = i_jbe
    i_jnae = i_jb
    i_jnb = i_jae
    i_jnbe = i_ja
    i_jnc = i_jae

    def i_jmp(self, op):
        tgt = self.getOperObj(op, 0)
        if not tgt.isDiscrete():
            # indirect jmp... table!

            return [(Const(tva, self._psize), eq(tgt, Const(tva, self._psize))) for fr, tva, tp, flag in self.vw.getXrefsFrom(op.va) if tp == REF_CODE]

    def i_jne(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_eq', self._psize)))

    i_jng = i_jle
    i_jnge = i_jl
    i_jnl = i_jge
    i_jnle = i_jg

    def i_jnp(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_pf', self._psize)))

    def i_jns(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_sf', self._psize)))

    i_jnz = i_jne

    def i_jp(self, op):
        return self._cond_jmp(op, Var('eflags_pf', self._psize))

    i_jpe = i_jp

    def i_jpo(self, op):
        return self._cond_jmp(op, cnot(Var('eflags_pf', self._psize)))

    def i_js(self, op):
        return self._cond_jmp(op, Var('eflags_sf', self._psize))

    def i_jecxz(self, op):
        return self._cond_jmp(op, cnot(Var('ecx', self._psize)))

    i_jz = i_je

    def i_lea(self, op):
        obj = self.getOperAddrObj(op, 1)
        self.setOperObj(op, 0, obj)

    def i_mov(self, op):
        o = self.getOperObj(op, 1)
        self.setOperObj(op, 0, o)

    # All of these mov's are technically different in i386/amd64, but really only differ by widths and sizes, and not by
    # actual mathematical operation
    i_movnti = i_mov
    i_movq = i_mov
    i_movd = i_mov
    i_movaps = i_mov
    i_movapd = i_mov
    i_movups = i_mov
    i_movdqu = i_mov
    i_movdqa = i_mov
    i_movd_q = i_mov

    def i_movlps(self, op):
        valu = self.getOperObj(op, 1) & self._sz_masks[8]
        self.setOperObj(op, 0, valu)

    def i_movhps(self, op):
        valu = (self.getOperObj(op, 1) & self._sz_masks[8]) << Const(64, self._psize)
        self.setOperObj(op, 0, valu)

    def i_vmovlps(self, op):
        if op.opers[0].isDeref():
            src = self.getOperObj(op, 1)
            self.setOperObj(op, 0, (src & self._sz_masks[8]))
        else:
            # 3 operands
            src1 = (self.getOperObj(op, 0, 1) & self._sz_masks[8]) << Const(64, self._psize)
            src2 = self.getOperObj(op, 0, 2) & self._sz_masks[8]
            self.setOperObj(op, 0, src1 | src2)

    def i_vmovhps(self, op):
        # which version of vmovhps are we looking at?
        if op.opers[0].isDeref():
            src = self.getOperObj(op, 1)
            self.setOperObj(op, 0, (src >> Const(64, self._psize)) & self._sz_masks[8])
        else:
            # 3 operands
            src1 = self.getOperObj(op, 0, 1) & self._sz_masks[8]
            src2 = (self.getOperObj(op, 0, 2) & self._sz_masks[8]) << Const(64, self._psize)
            self.setOperObj(op, 0, src1 | src2)

    def _movs(self, op, width=-1):
        si = Var(self.__srcp__, self._psize)
        di = Var(self.__destp__, self._psize)
        mem = Mem(si, Const(width, self._psize))
        self.effWriteMemory(di, Const(width, self._psize), mem)
        # This is a translation of if df == 0: add, else subtract that
        # the movs{b,w,d,q} are supposed to deal with
        mod = Const(width, self._psize) * (Const(1, self._psize) - Const(2, self._psize) * Var('eflags_df', self._psize))
        self.effSetVariable(self.__srcp__, si + mod)
        self.effSetVariable(self.__destp__, di + mod)

    def i_movsb(self, op):
        return self._movs(op, width=1)

    def i_movsd(self, op):
        # Unfortunately, these have the same mnemonic
        if op.opcode & i386_opconst.INS_STRMOV:
            # Since x86 doesn't differentiate movsw and movsd via opcode, we could try to infer it via the operators,
            # but those are for documention only (and even intel says those could be nothing but lies), so just go with
            # the default for now
            return self._movs(op, width=4)
        else:
            return self.i_mov(op)

    def i_movsx(self, op):
        dsize = op.opers[0].tsize
        ssize = op.opers[1].tsize
        v2 = o_sextend(self.getOperObj(op, 1), Const(ssize, self._psize))
        self.setOperObj(op, 0, v2)

    def i_movzx(self, op):
        v2 = self.getOperObj(op, 1)
        self.setOperObj(op, 0, v2)

    def i_neg(self, op):
        obj = self.getOperObj(op, 0)
        width = obj.getWidth()
        self.setOperObj(op, 0, o_sub(Const(0, width), obj, width))

    def i_nop(self, op):
        pass

    def i_prefetch(self, op):
        pass

    def i_prefetchw(self, op):
        pass

    def i_pmovmskb(self, op):
        v2 = self.getOperObj(op, 1)
        res = Const(0, self._psize)
        for i in range(v2.getWidth()):
            i2 = Const(i, self._psize)
            mask = Const(1 << (7 + i*8), self._psize)
            res |= (v2 & mask) << i2
        self.setOperObj(op, 0, res)

    i_vpmovmskb = i_pmovmskb

    def i_not(self, op):
        v1 = self.getOperObj(op, 0)
        v1width = v1.getWidth()
        v1 ^= Const(e_bits.u_maxes[v1width], v1width)
        self.setOperObj(op, 0, v1)

    def i_or(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        obj = o_or(v1, v2, v1.getWidth())

        self.effSetVariable('eflags_gt', gt(v1, v2))
        self.effSetVariable('eflags_lt', lt(v1, v2))
        self.effSetVariable('eflags_of', Const(0, self._psize))
        self.effSetVariable('eflags_cf', Const(0, self._psize))
        self.effSetVariable('eflags_sf', lt(obj, Const(0, self._psize))) # v1 | v2 < 0
        self.effSetVariable('eflags_eq', eq(obj, Const(0, self._psize))) # v1 & v2 == 0
        self.setOperObj(op, 0, obj)

    def i_pand(self, op, off=0):
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off+1)

        obj = o_and(v1, v2, v1.getWidth())

        self.setOperObj(op, 0, obj)

    def i_pandn(self, op, off=0):
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off+1)

        v1 = self.invert(v1)
        obj = o_and(v1, v2, v1.getWidth())

        self.setOperObj(op, 0, obj)

    def i_vpand(self, op):
        self.i_pand(op, off=1)

    def i_vpandn(self, op):
        self.i_pandn(op, off=1)

    # TODO: pshufb has a conditional in it where it only sets it if the high bit of the src byte
    # its chosen is set. How do we model that in symboliks?
    # shift the msb down, size it to 1 and the sign extend it and use that as a mask?
    #def i_pshufb(self, op):
        #pass

    def i_pshufd(self, op):
        dst = self.getOperObj(op, 0)
        src = self.getOperObj(op, 1)
        ordr = self.getOperObj(op, 2)

        res = Const(0, self._psize)
        mask = Const((2 ** 32) - 1, self._psize)
        scal = Const(32, self._psize)

        for i in range(dst.getWidth()):
            indx = (ordr >> Const(2*i, self._psize)) & Const(0x3, self._psize)
            res |= ((src >> (indx * scal)) & mask) << Const(i * 32, self._psize)

        self.setOperObj(op, 0, res)

    # i_vpshufd = i_pshufd

    def i_pshuflw(self, op, offset=0):
        dst = self.getOperObj(op, 0)
        src = self.getOperObj(op, 1)
        ordr = self.getOperObj(op, 2)

        mask = self._sz_masks[2] << Const(offset, self._psize)
        clear = self._sz_masks[8] << Const(64 - offset, self._psize)
        ordmask = Const(3, self._psize)

        res = src & clear

        for i in range(4):
            ordshft = Const(2 * i, self._psize)
            indx = (ordr >> ordshft) & ordmask
            indx = indx * Const(16, self._psize)
            res |= ((src >> indx) & mask) << Const(i * 16, self._psize)

        if dst.getWidth() == 32:
            upoff = Const(128, self._psize)
            src >>= upoff
            res |= (src & clear) << upoff
            for i in range(4):
                ordshft = Const(2 * i, self._psize)
                indx = (ordr >> ordshft) & ordmask
                indx = indx * Const(16, self._psize)
                valu = ((src >> indx) >> offset) & mask
                res |= valu << Const((i * 16) + 128 + offset, self._psize)

        self.setOperObj(op, 0, res)

    i_vpshuflw = i_pshuflw

    def i_pshufhw(self, op):
        self.i_pshuflw(op, offset=64)

    i_vpshufhw = i_pshufhw

    def i_push(self, op):
        v1 = self.getOperObj(op, 0)
        sp = self.getRegObj(self._reg_ctx._rctx_spindex)
        self.effSetVariable(self.__sp__, sp - Const(self._psize, self._psize))
        # NOTE: we do the write after the set, so there is no need
        # to write to sp - psize...
        self.effWriteMemory(Var(self.__sp__, self._psize), Const(self._psize, self._psize), v1)

    def i_pop(self, op):
        self.setOperObj(op, 0, Mem(Var(self.__sp__, self._psize), Const(self._psize, self._psize)))
        self.effSetVariable(self.__sp__, Var(self.__sp__, self._psize) + Const(self._psize, self._psize))

    def _simdshift(self, op, func, width, off=0):
        res = Const(0, self._psize)
        valu = self.getOperObj(op, off)
        count = self.getOperObj(op, off + 1)
        if count.isDiscrete():
            if count.solve() >= width:
                self.setOperObj(op, 0, Const(0, self._psize))
                return

        # TODO: Pre-gen these?
        mask = Const((2 ** width) - 1, valu.getWidth())
        iters = int(valu.getWidth() / width)
        bitCount = count * Const(8, self._psize)
        for i in range(iters):
            shift = Const(i * width * 8, self._psize)
            tmp = (valu >> shift) & mask
            res |= func(tmp, bitCount) << shift

        self.setOperObj(op, 0, res)

    def i_psrlw(self, op):
        self._simdshift(op, operator.rshift, 2)

    def i_psrld(self, op):
        self._simdshift(op, operator.rshift, 4)

    def i_psrlq(self, op):
        self._simdshift(op, operator.rshift, 8)

    def i_psrldq(self, op):
        self._simdshift(op, operator.rshift, 16)

    def i_vpsrlw(self, op):
        self._simdshift(op, operator.rshift, 2, off=1)

    def i_vpsrld(self, op):
        self._simdshift(op, operator.rshift, 4, off=1)

    def i_vpsrlq(self, op):
        self._simdshift(op, operator.rshift, 8, off=1)

    def i_vpsrldq(self, op):
        self._simdshift(op, operator.rshift, 16, off=1)

    def i_psllw(self, op):
        self._simdshift(op, operator.lshift, 2)

    def i_pslld(self, op):
        self._simdshift(op, operator.lshift, 4)

    def i_psllq(self, op):
        self._simdshift(op, operator.lshift, 8)

    def i_pslldq(self, op):
        self._simdshift(op, operator.lshift, 16)

    def i_vpsllw(self, op):
        self._simdshift(op, operator.lshift, 2, off=1)

    def i_vpslld(self, op):
        self._simdshift(op, operator.lshift, 4, off=1)

    def i_vpsllq(self, op):
        self._simdshift(op, operator.lshift, 8, off=1)

    def i_vpslldq(self, op):
        self._simdshift(op, operator.lshift, 16, off=1)

    def i_pxor(self, op, off=0):
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off+1)
        obj = o_xor(v1, v2, v1.getWidth())
        self.setOperObj(op, 0, obj)

    def i_vpxor(self, op):
        self.i_pxor(op, off=1)

    def i_por(self, op, off=0):
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off+1)
        obj = o_or(v1, v2, v1.getWidth())
        self.setOperObj(op, 0, obj)

    def i_vpor(self, op):
        self.i_por(op, off=1)

    def _parith(self, op, func, width, off=0):
        res = Const(0, self._psize)
        dst = self.getOperObj(op, off)
        src = self.getOperObj(op, off+1)
        bitwidth = width * 8

        mask = Const((2 ** bitwidth) - 1, dst.getWidth())
        iters = int(dst.getWidth() / width)
        for i in range(iters):
            shift = Const(i * bitwidth, self._psize)
            opA = (dst >> shift) & mask
            opB = (src >> shift) & mask
            res |= (func(opA, opB) & mask) << shift

        self.setOperObj(op, 0, res)

    def i_paddb(self, op, off=0):
        self._parith(op, operator.add, 1)

    def i_paddw(self, op):
        self._parith(op, operator.add, 2)

    def i_paddd(self, op):
        self._parith(op, operator.add, 4)

    def i_paddq(self, op):
        self._parith(op, operator.add, 8)

    def i_vpaddb(self, op, off=0):
        self._parith(op, operator.add, 1, off=1)

    def i_vpaddw(self, op):
        self._parith(op, operator.add, 2, off=1)

    def i_vpaddd(self, op):
        self._parith(op, operator.add, 4, off=1)

    def i_vpaddq(self, op):
        self._parith(op, operator.add, 8, off=1)

    def i_psubb(self, op):
        self._parith(op, operator.sub, 1)

    def i_psubw(self, op):
        self._parith(op, operator.sub, 2)

    def i_psubd(self, op):
        self._parith(op, operator.sub, 4)

    def i_psubq(self, op):
        self._parith(op, operator.sub, 8)

    def i_vpsubb(self, op):
        self._parith(op, operator.sub, 1, off=1)

    def i_vpsubw(self, op):
        self._parith(op, operator.sub, 2, off=1)

    def i_vpsubd(self, op):
        self._parith(op, operator.sub, 4, off=1)

    def i_vpsubq(self, op):
        self._parith(op, operator.sub, 8, off=1)

    def i_ret(self, op):
        self.effSetVariable(self.__ip__, Mem(Var(self.__sp__, self._psize), Const(self._psize, self._psize)))
        sp = Var(self.__sp__, self._psize)
        sp += Const(self._psize, self._psize)
        if len(op.opers):
            sp += self.getOperObj(op, 0)
        self.effSetVariable(self.__sp__, sp)

    # TODO: We really should condense these four into one megafunction
    def i_rol(self, op):
        valu = self.getOperObj(op, 0)
        shft = self.getOperObj(op, 1)

        dsize = op.opers[0].tsize
        shft = shft % Const(dsize * 8, self._psize)

        shft_in = valu >> (Const(dsize * 8, self._psize) - shft)

        res = valu << shft
        res |= shft_in

        self.setOperObj(op, 0, res)

    def i_ror(self, op):
        valu = self.getOperObj(op, 0)
        shft = self.getOperObj(op, 1)

        dsize = op.opers[0].tsize
        shft = shft % Const(dsize * 8, self._psize)

        shft_in = valu << (Const(dsize * 8, self._psize) - shft)

        res = valu >> shft
        res |= shft_in

        self.setOperObj(op, 0, res)

    def i_rcr(self, op):
        valu = self.getOperObj(op, 0)
        shft = self.getOperObj(op, 1)

        dsize = op.opers[0].tsize
        bitsize = Const(dsize * 8, self._psize)
        shft = shft % bitsize
        tot = (Var('eflags_cf', self._psize) << bitsize) | valu

        shft_in = tot << (bitsize - shft)
        tot >>= shft
        tot |= shft_in

        cf = tot >> (bitsize - Const(1, dsize))
        self.setOperObj(op, 0, tot)
        self.effSetVariable('eflags_cf', cf)

    def i_rcl(self, op):
        valu = self.getOperObj(op, 0)
        shft = self.getOperObj(op, 1)

        dsize = op.opers[0].tsize
        bitsize = Const(dsize * 8, self._psize)
        shft = shft % bitsize
        tot = (Var('eflags_cf', self._psize) << bitsize) | valu

        shft_in = tot >> (bitsize - shft)

        tot <<= shft
        tot |= shft_in

        cf = tot >> (bitsize - Const(1, dsize))
        self.setOperObj(op, 0, tot)
        self.effSetVariable('eflags_cf', cf)

    def i_sar(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)

        # Arithmetic shifts are multiplies and divides!
        res = o_div(v1, o_pow(Const(2, self._psize), v2, self._psize), v1.getWidth())

        if v2.isDiscrete():
            bit = v2.solve()
            if bit == 1:
                self.effSetVariable('eflags_of', Const(0, self._psize))
            if bit != 0:
                power = o_pow(Const(2, self._psize), Const(bit, self._psize), self._psize)
                cf = o_div(v1 & power, power, v1.getWidth())
                self.effSetVariable('eflags_cf', cf)
        # TODO: I'm hesitant to put this in here blindly, because on a shift of 0, we'd generate an effect
        # that doesn't happen
        #else:
        #    power = o_pow(Const(2, self._psize), v2, self._psize)
        #    cf = o_div(v1 & power, power, v1.getWidth())
        #    cf = v1 & (o_pow(2, self._psize), v2, self._psize - 1)

        self.effSetVariable('eflags_gt', gt(v1, v2))
        self.effSetVariable('eflags_lt', lt(v1, v2))
        self.effSetVariable('eflags_sf', lt(res, Const(0, self._psize)))  # v1 | v2 < 0
        self.effSetVariable('eflags_eq', eq(res, Const(0, self._psize)))  # v1 & v2 == 0

        self.setOperObj(op, 0, res)

    def i_sbb(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        obj = o_sub(v1, v2, v1.getWidth())  # FIXME borrow!
        self.effSetVariable('eflags_gt', gt(v1, v2))  # v1 - v2 > 0 :: v1 > v2
        self.effSetVariable('eflags_lt', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_sf', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_eq', eq(v1, v2))  # v1 - v2 == 0 :: v1 == v2
        self.setOperObj(op, 0, v1 - v2)

    def i_cwde(self, op):
        v1 = o_sextend(self.getRegObj(e_i386.REG_AX), Const(self._psize, self._psize))
        self.setRegByName('eax', v1)

    def _carry_eq(self, x):
        return eq(Var('eflags_cf', self._psize), Const(x, self._psize))

    def _zero_eq(self, x):
        return eq(Var('eflags_eq', self._psize), Const(x, self._psize))

    def _overflow_eq(self, x):
        return eq(Var('eflags_eq', self._psize), Const(x, self._psize))

    def _signed_eq(self, x):
        return eq(Var('eflags_eq', self._psize), Const(x, self._psize))

    def _parity_eq(self, x):
        return eq(Var('eflags_pf', self._psize), Const(x, self._psize))

    def i_seta(self, op):
        self.setOperObj(op, 0, self._carry_eq(0) & self._zero_eq(0))

    def i_setae(self, op):
        self.setOperObj(op, 0, self._carry_eq(0))

    def i_setb(self, op):
        self.setOperObj(op, 0, self._carry_eq(1))

    def i_setbe(self, op):
        self.setOperObj(op, 0, self._zero_eq(1) | self._carry_eq(1))

    i_setc = i_setb

    def i_sete(self, op):
        self.setOperObj(op, 0, self._zero_eq(1))

    def i_setg(self, op):
        signed = eq(Var('eflags_sf', self._psize), Var('eflags_of', self._psize))
        self.setOperObj(op, 0, self._overflow_eq(0) & signed)

    def i_setge(self, op):
        signed = eq(Var('eflags_sf', self._psize), Var('eflags_of', self._psize))
        self.setOperObj(op, 0, signed)

    def i_setl(self, op):
        not_signed = ne(Var('eflags_sf', self._psize), Var('eflags_of', self._psize))
        self.setOperObj(op, 0, not_signed)

    def i_setle(self, op):
        equal = eq(Var('eflags_eq', self._psize), Const(1, self._psize))
        not_signed = ne(Var('eflags_sf', self._psize), Var('eflags_of', self._psize))
        self.setOperObj(op, 0, self._zero_eq(1) | not_signed)

    i_setna = i_setbe

    def i_setnae(self, op):
        self.setOperObj(op, 0, self._carry_eq(1))

    def i_setnb(self, op):
        self.setOperObj(op, 0, self._carry_eq(0))

    i_setnbe = i_seta
    i_setnc = i_setnae

    def i_setne(self, op):
        self.setOperObj(op, 0, self._zero_eq(0))

    i_setng = i_setle
    i_setnge = i_setl
    i_setnl = i_setge
    i_setnle = i_setg

    def i_setno(self, op):
        self.setOperObj(op, 0, self._overflow_eq(0))

    def i_setnp(self, op):
        self.setOperObj(op, 0, ne(self._parity_eq(1)))
    i_setpo = i_setnp

    def i_setns(self, op):
        self.setOperObj(op, 0, self._signed_eq(0))

    i_setnz = i_setne

    def i_seto(self, op):
        self.setOperObj(op, 0, self._overflow_eq(0))

    def i_setp(self, op):
        self.setOperObj(op, 0, self._parity_eq(1))

    i_setpe = i_setp

    def i_sets(self, op):
        self.setOperObj(op, 0, self._signed_eq(1))
    i_setz = i_sete

    def i_shlx(self, op):
        v1 = self.getOperObj(op, 1)
        v2 = self.getOperObj(op, 2)
        res = v1 << v2
        self.setOperObj(op, 0, res)

    def i_shrx(self, op):
        v1 = self.getOperObj(op, 1)
        v2 = self.getOperObj(op, 2)
        res = v1 >> v2
        self.setOperObj(op, 0, res)

    def i_sarx(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        res = o_div(v1, o_pow(Const(2, self._psize), v2, self._psize), v1.getWidth())
        self.setOperObj(op, 0, res)

    def i_shl(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        zero = Const(0, self._psize)

        #if v2.isDiscrete() and v2.solve() > 0xff:
            #v2 = v2 & 0xff

        # No effect (not even flags) if shift is 0
        if v2.isDiscrete() and v2.solve() == 0:
            return

        res = v1 << v2
        self.setOperObj(op, 0, res)

        self.effSetVariable('eflags_gt', gt(res, zero))
        self.effSetVariable('eflags_lt', lt(res, zero))
        self.effSetVariable('eflags_sf', lt(res, zero))
        self.effSetVariable('eflags_eq', eq(res, zero))

    def i_shr(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        zero = Const(0, self._psize)

        # No effect (not even flags) if shift is 0
        if v2.isDiscrete():
            bit = v2.solve()
            if bit == 0:
                return
            elif bit == 1:
                tsize = v1.getWidth()
                power = o_pow(Const(2, tsize), Const(tsize - 1, self._psize), self._psize)
                msb = o_div((v1 & power), power, v1.getWidth())
                self.effSetVariable('eflags_of', msb)

        res = v1 >> v2
        self.setOperObj(op, 0, res)
        self.effSetVariable('eflags_gt', gt(res, zero))
        self.effSetVariable('eflags_lt', lt(res, zero))
        self.effSetVariable('eflags_sf', lt(res, zero))
        self.effSetVariable('eflags_eq', eq(res, zero))

    def i_std(self, op):
        self.effSetVariable('eflags_df', Const(1, self._psize))

    def i_sub(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        obj = o_sub(v1, v2, v1.getWidth())
        dsize = op.opers[0].tsize
        self.effSetVariable('eflags_gt', gt(v1, v2))  # v1 - v2 > 0 :: v1 > v2
        self.effSetVariable('eflags_lt', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_sf', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_eq', eq(v1, v2))  # v1 - v2 == 0 :: v1 == v2
        self.effSetVariable('eflags_pf', self._generate_parity(obj))

        f = gt(obj, Const(e_bits.s_maxes[dsize>>1], dsize))
        self.effSetVariable('eflags_of', f)
        self.setOperObj(op, 0, obj)

    def i_subsd(self, op, off=0):
        '''
        None of the ref docs say subsd affects any flags
        '''
        v1 = self.getOperObj(op, off)
        v2 = self.getOperObj(op, off+1)
        obj = o_sub(v1, v2, v1.getWidth())
        self.setOperObj(op, 0, obj)

    i_subss = i_subsd

    def i_vsubsd(self, op):
        self.i_subsd(op, off=1)
    i_vsubss = i_vsubsd

    def i_test(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        obj = o_and(v1, v2, v1.getWidth())
        zero = Const(0, self._psize)

        self.effSetVariable('eflags_gt', gt(v1, v2))
        self.effSetVariable('eflags_lt', lt(obj, zero))  # ( SF != OF ) ( OF is cleared )
        self.effSetVariable('eflags_sf', lt(obj, zero))  # v1 & v2 < 0
        self.effSetVariable('eflags_eq', eq(obj, zero))  # v1 & v2 == 0

        self.effSetVariable('eflags_of', zero)
        self.effSetVariable('eflags_cf', zero)

        pf = o_xor(o_xor(v1, v2, v1.getWidth()), Const(-1, self._psize), self._psize)
        self.effSetVariable('eflags_pf', pf)

    def i_xadd(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        obj = o_add(v1, v2, v1.getWidth())
        self.setOperObj(op, 0, obj)
        self.setOperObj(op, 1, v1)
        self.effSetVariable('eflags_gt', gt(v1, v2))
        self.effSetVariable('eflags_lt', lt(v1, v2))
        self.effSetVariable('eflags_sf', lt(obj, Const(0, self._psize)))
        self.effSetVariable('eflags_eq', eq(obj, Const(0, self._psize)))
        dsize = op.opers[0].tsize
        f = gt(obj, Const(e_bits.s_maxes[dsize>>1], dsize))
        self.effSetVariable('eflags_of', f)
        self.effSetVariable('eflags_pf', self._generate_parity(obj))

    def i_xor(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        obj = o_xor(v1, v2, v1.getWidth())
        zero = Const(0, self._psize)

        self.effSetVariable('eflags_gt', gt(v1, v2))
        self.effSetVariable('eflags_lt', lt(v1, v2))
        self.effSetVariable('eflags_sf', lt(obj, zero))  # v1 ^ v2 < 0
        self.effSetVariable('eflags_eq', eq(obj, zero))  # v1 & v2 == 0
        self.effSetVariable('eflags_cf', zero)
        self.effSetVariable('eflags_of', zero)
        self.effSetVariable('eflags_pf', self._generate_parity(obj))
        self.setOperObj(op, 0, obj)

    def i_xorpd(self, op):
        v1 = self.getOperObj(op, 0)
        v2 = self.getOperObj(op, 1)
        obj = o_xor(v1, v2, v1.getWidth())
        self.setOperObj(op, 0, obj)

    i_xorps = i_xorpd

    def i_sqrtsd(self, op):
        v1 = self.getOperObj(op, 1)
        obj = o_pow(v1, Const(0.5, self._psize))
        self.setOperObj(op, 0, obj)
    i_sqrtss = i_sqrtsd

    def i_cmpxchg(self, op):

        # FIXME CATASTROPHIC THIS CONTAINS BRANCHING LOGIC STATE!
        # FOR NOW WE JUST DO IT WITHOUT ANY CONDITIONAL
        x = self.getOperObj(op, 0)
        y = self.getOperObj(op, 1)
        self.effSetVariable('i386_xchg_tmp', x)
        self.setOperObj(op, 0, y)
        self.setOperObj(op, 1, Var('i386_xchg_tmp', self._psize))

    def i_xchg(self, op):
        # NOTE: requires using temp var because each asignment occurs
        # seperately. (even though the API makes it look like you've
        # got your own copy... ;) )
        x = self.getOperObj(op, 0)
        y = self.getOperObj(op, 1)
        self.effSetVariable('i386_xchg_tmp', x)
        self.setOperObj(op, 0, y)
        self.setOperObj(op, 1, Var('i386_xchg_tmp', self._psize))

    def i_rdtsc(self, op):
        self.setRegByName('edx', Var('TSC_HIGH', self._psize))
        self.setRegByName('eax', Var('TSC_LOW', self._psize))

    def i_leave(self, op):
        self.effSetVariable(self.__sp__, Var(self.__bp__, self._psize))
        self.effSetVariable(self.__bp__, Mem(Var(self.__sp__, self._psize), Const(self._psize, self._psize)))
        self.effSetVariable(self.__sp__, o_add(Var(self.__sp__, self._psize), Const(self._psize, self._psize), self._psize))

    def i_pextrb(self, op, width=1):
        dst = self.getOperObj(op, 0)
        src = self.getOperObj(op, 1)
        select = self.getOperObj(op, 2)

        src >>= select * Const(width * 8, self._psize)
        src &= self._sz_masks[width]
        if width != 1:
            dst &= Const(~((2**(8*width)) - 1), self._psize)
            src |= dst

        self.setOperObj(op, 0, src)

    def i_pextrw(self, op):
        self.i_pextrb(op, width=2)

    def i_pextrd(self, op):
        self.i_pextrb(op, width=4)

    def i_stosb(self, op, width=1):
        # FIXME omg segments in symboliks?
        # base, size = self.segments[SEG_ES]
        di = Var(self.__destp__, self._psize)
        mod = Const(width, self._psize) * (Const(1, self._psize) - Const(2, self._psize) * Var('eflags_df', self._psize))
        self.effWriteMemory(di, Const(self._psize, self._psize), Var('eax', self._psize))
        self.effSetVariable(self.__destp__, di + mod)

    def i_stosd(self, op):
        return self.i_stosb(op, width=4)

    def _cmps(self, op, width=-1):
        si = Var(self.__srcp__, self._psize)
        di = Var(self.__destp__, self._psize)
        v1 = Mem(si, Const(width, self._psize))
        v2 = Mem(di, Const(width, self._psize))
        self.effSetVariable('eflags_cf', gt(v2, v1))  #
        self.effSetVariable('eflags_gt', gt(v1, v2))  # v1 - v2 > 0 :: v1 > v2
        self.effSetVariable('eflags_lt', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_sf', lt(v1, v2))  # v1 - v2 < 0 :: v1 < v2
        self.effSetVariable('eflags_eq', eq(v1, v2))  # v1 - v2 == 0 :: v1 == v2

    def i_cmpsb(self, op):
        return self._cmps(op, width=1)

    def i_cmpsd(self, op):
        return self._cmps(op, width=4)

class i386SymbolikTranslator(IntelSymbolikTranslator):
    __arch__ = e_i386.i386Module
    __ip__ = 'eip' # we could use regctx.getRegisterName if we want.
    __sp__ = 'esp' # we could use regctx.getRegisterName if we want.
    __bp__ = 'ebp' # we could use regctx.getRegisterName if we want.
    __srcp__ = 'esi'
    __destp__ = 'edi'

    def i_pushad(self, op):
        esp = self.getRegObj(e_i386.REG_ESP)
        self.effWriteMemory(esp - Const(4, self._psize), Const(4, self._psize), self.getRegObj(e_i386.REG_EAX))
        self.effWriteMemory(esp - Const(8, self._psize), Const(4, self._psize), self.getRegObj(e_i386.REG_ECX))
        self.effWriteMemory(esp - Const(12, self._psize), Const(4, self._psize), self.getRegObj(e_i386.REG_EDX))
        self.effWriteMemory(esp - Const(16, self._psize), Const(4, self._psize), self.getRegObj(e_i386.REG_EBX))
        self.effWriteMemory(esp - Const(20, self._psize), Const(4, self._psize), esp)
        self.effWriteMemory(esp - Const(24, self._psize), Const(4, self._psize), self.getRegObj(e_i386.REG_EBP))
        self.effWriteMemory(esp - Const(28, self._psize), Const(4, self._psize), self.getRegObj(e_i386.REG_ESI))
        self.effWriteMemory(esp - Const(32, self._psize), Const(4, self._psize), self.getRegObj(e_i386.REG_EDI))
        # FIXME re-order these!
        self.effSetVariable('esp', esp - Const(32, self._psize))

    def i_pushfd(self, op):
        esp = self.getRegObj(e_i386.REG_ESP)
        eflags = self.getRegObj(e_i386.REG_EFLAGS)
        self.effSetVariable('esp', esp - Const(4, self._psize))
        self.effWriteMemory(Var('esp', self._psize), Const(4, self._psize), eflags)

    def i_popad(self, op):
        esp = self.getRegObj(e_i386.REG_ESP)
        self.effSetVariable('edi', Mem(esp, Const(4, self._psize)))
        self.effSetVariable('esi', Mem(esp + Const(4, self._psize), Const(4, self._psize)))
        self.effSetVariable('ebp', Mem(esp + Const(8, self._psize), Const(4, self._psize)))
        self.effSetVariable('ebx', Mem(esp + Const(16, self._psize), Const(4, self._psize)))
        self.effSetVariable('edx', Mem(esp + Const(20, self._psize), Const(4, self._psize)))
        self.effSetVariable('ecx', Mem(esp + Const(24, self._psize), Const(4, self._psize)))
        self.effSetVariable('eax', Mem(esp + Const(28, self._psize), Const(4, self._psize)))
        self.effSetVariable('esp', esp + Const(32, self._psize))

class i386ArgDefSymEmu(vsym_callconv.ArgDefSymEmu):
    __xlator__ = i386SymbolikTranslator

class i386SymCallingConv(vsym_callconv.SymbolikCallingConvention):
    __argdefemu__ = i386ArgDefSymEmu

    def __init__(self, xlator=None):
        super().__init__(xlator=xlator)

class StdCall(i386SymCallingConv, e_i386.StdCall):
    pass

class Cdecl(i386SymCallingConv, e_i386.Cdecl):
    pass

class ThisCall(i386SymCallingConv, e_i386.ThisCall):
    pass

class MsFastCall(i386SymCallingConv, e_i386.MsFastCall):
    pass

class BFastCall(i386SymCallingConv, e_i386.BFastCall):
    pass

class ThisCall_Caller(i386SymCallingConv, e_i386.ThisCall_Caller):
    pass

class MsFastCall_Caller(i386SymCallingConv, e_i386.MsFastCall_Caller):
    pass

class BFastCall_Caller(i386SymCallingConv, e_i386.BFastCall_Caller):
    pass

class i386SymFuncEmu(vsym_analysis.SymbolikFunctionEmulator):

    __width__ = 4

    def __init__(self, vw, *args, xlator=None):
        vsym_analysis.SymbolikFunctionEmulator.__init__(self, vw, *args, xlator=xlator)
        self.setStackBase(0xbfbff000, 16384)

        self.addCallingConvention('cdecl', Cdecl(xlator=xlator))
        self.addCallingConvention('stdcall', StdCall(xlator=xlator))
        self.addCallingConvention('thiscall', ThisCall(xlator=xlator))
        self.addCallingConvention('bfastcall', BFastCall(xlator=xlator))
        self.addCallingConvention('msfastcall', MsFastCall(xlator=xlator))

        self.addCallingConvention('thiscall_caller', ThisCall_Caller(xlator=xlator))
        self.addCallingConvention('bfastcall_caller', BFastCall_Caller(xlator=xlator))
        self.addCallingConvention('msfastcall_caller', MsFastCall_Caller(xlator=xlator))
        # FIXME possibly decide this by platform/format?
        self.addCallingConvention(None, Cdecl(xlator=xlator))

        self.addFunctionCallback('ntdll.eh_prolog', self._eh_prolog)
        self.addFunctionCallback('ntdll.seh3_prolog', self._seh3_prolog)
        self.addFunctionCallback('ntdll.seh3_epilog', self._seh3_epilog)
        self.addFunctionCallback('ntdll._alloca_probe', self.alloca_probe)

        #self.writeSymMemory( Mem(Var('fs') + 292, 4)

    def getStackCounter(self):
        return self.getSymVariable('esp')

    def setStackCounter(self, symobj):
        self.setSymVariable('esp', symobj)

    def _eh_prolog(self, emu, fname, argv):

        # swap out [ esp ] (saved eip) to ebp
        # and set ebp to current esp (std frame)
        ebp = emu.getSymVariable('ebp')
        esp = emu.getSymVariable('esp')
        eax = emu.getSymVariable('eax')
        emu.writeSymMemory(esp, ebp)
        emu.setSymVariable('ebp', esp)

        # now carry out 3 symbolik pushes
        esp -= Const(4, 4)
        emu.writeSymMemory(esp, Var('eh_ffffffff',4))
        esp -= Const(4, 4)
        emu.writeSymMemory(esp, eax)
        esp -= Const(4, 4)
        emu.writeSymMemory(esp, Var('eh_c0c0c0c0',4))
        self.setSymVariable('esp', esp)

        return eax

    def alloca_probe(self, emu, fname, argv):
        esp = emu.getSymVariable('esp')
        eax = emu.getSymVariable('eax')
        # Update the stack size if eax is discrete
        if eax.isDiscrete(emu=emu):
            stackadd = eax.solve(emu=emu)
            stackadd += 4096 - (stackadd % 4096)
            stacksize = emu.getStackSize()
            emu.setStackSize( stacksize + stackadd )

        emu.setSymVariable('esp', esp-eax)
        #if eax < 0x1000:
            #eax -= Const(4, self.__width__)
            #emu.setSymVariable('esp', esp-eax)
        #else:
            #while eax > 0x1000:
                #eax -= Const(0x1000, self.__width__)
                #emu.setSymVariable('esp', esp-Const(0x1000, self.__width__))
                #esp -= Const(0x1000, self.__width__)
            #emu.setSymVariable('esp', esp-eax)

    def _seh3_prolog(self, emu, fname, argv):

        scopetable, localsize = argv
        esp = emu.getSymVariable('esp')

        emu.writeSymMemory(esp + Const(4, self.__width__),  emu.getSymVariable('ebp'))
        emu.writeSymMemory(esp, scopetable)
        emu.writeSymMemory(esp - Const(4, self.__width__), Var('ntdll.seh3_handler', 4))     # push
        emu.writeSymMemory(esp - Const(8, self.__width__), Var('saved_seh3_scopetable', 4))  # push
        emu.writeSymMemory(esp - Const(12, self.__width__), emu.getSymVariable('ebx'))
        emu.writeSymMemory(esp - Const(16, self.__width__), emu.getSymVariable('esi'))
        emu.writeSymMemory(esp - Const(20, self.__width__), emu.getSymVariable('edi'))
        emu.setSymVariable('esp', (esp - Const(20, self.__width__)) - localsize)
        emu.setSymVariable('ebp', (esp + Const(4, self.__width__)))

        return scopetable

    def _seh3_epilog(self, emu, fname, argv):
        edi = emu.getSymVariable('edi')
        esi = emu.getSymVariable('esi')
        ebx = emu.getSymVariable('ebx')

        esp = emu.getSymVariable('esp')
        # FIXME do seh restore...
        emu.setSymVariable('edi', edi)
        emu.setSymVariable('esi', esi)
        emu.setSymVariable('ebx', ebx)
        ebp = emu.getSymVariable('ebp')
        savedbp = emu.readSymMemory(ebp, Const(4, self.__width__))
        emu.setSymVariable('ebp', savedbp)
        emu.setSymVariable('esp', ebp + Const(4, self.__width__))
        return emu.getSymVariable('eax')

class i386SymbolikAnalysisContext(vsym_analysis.SymbolikAnalysisContext):
    __xlator__ = i386SymbolikTranslator
    __emu__ = i386SymFuncEmu



# APIs for Windows 64-bit ole32 library.
# Built as a delta from the 32-bit version.
# Format:  rettype, retname, callconv, exactname, arglist(type, name)
#          arglist type is one of ['int', 'void *']
#          arglist name is one of [None, 'funcptr', 'obj', 'ptr']

# List the normalized name of any 32-bit functions to omit.
api_32_omits = []

# Define any functions specific to 64-bit.
api_64_adds = {
    }


# Build from the 32-bit API, skipping omits, changing the calling convention,
# and adding any specific 64-bit functions.
api_defs_64 = {}

import vivisect.impapi.windows.ole_32 as m32
for normname, (rtype, rname, cconv, cname, cargs) in m32.api_defs.items():
    if normname in api_32_omits:
        continue
    api_defs_64[normname] = (rtype, rname, 'msx64call', cname, cargs)
api_defs_64.update(api_64_adds)



import os
import time
import logging
import unittest
import contextlib

import cobra
import cobra.dcode
import cobra.remoteapp

from subprocess import Popen, PIPE

logger = logging.getLogger(__name__)


class Foo:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def add(self):
        return self.x + self.y

    def mult(self):
        return self.x * self.y

    def pow(self):
        return self.x ** self.y


@contextlib.contextmanager
def getDcodeDmon():
    dmon = cobra.CobraDaemon(host='localhost', port=41923, msgpack=True)
    dmon.fireThread()
    cobra.dcode.enableDcodeServer(daemon=dmon)
    name = cobra.remoteapp.shareRemoteApp('cobra.testclient', appsrv=Foo(3, 4), daemon=dmon)
    yield (name, dmon)
    dmon.unshareObject(name)


def buildCobra(host, port, name):
    builder = cobra.initSocketBuilder(host, port)
    builder.setTimeout(5)
    return cobra.CobraProxy('cobra://%s:%s/%s?msgpack=1' % (host, port, name))


class CobraDcodeTest(unittest.TestCase):

    def test_cobra_dcode(self):
        with getDcodeDmon() as (name, dmon):
            # dmon.fireThread()
            srv = buildCobra(dmon.host, dmon.port, name)
            self.assertEqual(srv.add(), 7)
            self.assertEqual(srv.mult(), 12)
            self.assertEqual(srv.pow(), 81)


    def test_dcode_remotecode(self):
        '''
        This test can be confusing, because we're using Dcode to access the
        local machine, not a remote one.

        First, we setup a dcode server with a special base path of "cobra".
        This makes the contents of Vivisect's `cobra` directory accessible
        as a top-level.

        Next, we connect to the the Dcode server with the unittest process
        and import dcode from the "remote" cobra path offered by Dcode.
        
        Assertion is just a constant from the remote cobra module
        '''

        # setup server-side with special path: importing "cobra" as the base,
        # so anything in that directory becomes top-level, eg. cobra.dcode can
        # be imported as "import dcode")
        p = Popen("python3 -m cobra.dcode -P 12347 cobra".split(' '),
                executable='python3', stdin=PIPE, stderr=PIPE)

        buf = b''
        starttime = time.time()
        while b'Adding path "cobra" to system path' not in buf:
            # timeout after 10 seconds waiting for this string
            if time.time() - starttime > 10:
                break

            # peek allows us to check for data without blocking
            if len(p.stderr.peek()):
                buf += p.stderr.read1()

        time.sleep(.1)
        # setup client-side
        cobra.dcode.addDcodeServer('localhost', 12347)

        # this doesn't exist in the local PYTHONPATH, but should come from the Dcode server
        import dcode    # through Dcode, this is accessing cobra.dcode

        # now we'll test the constant COBRA_CALL (which we get through Dcode)
        self.assertEqual(dcode.cobra.COBRA_CALL, 1)

        # cleanup
        p.kill()
        p.wait()

